from abc import abstractmethod

import torch
import torch.nn as nn

from torch_mist.utils.caching import (
    reset_cache_before_call,
    reset_cache_after_call,
)


class MIEstimator(nn.Module):
    lower_bound: bool = False
    upper_bound: bool = False
    infomax_gradient: bool = False

    @abstractmethod
    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    @reset_cache_after_call
    def mutual_information(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        log_ratio = self.log_ratio(x, y)
        assert (
            not isinstance(x, torch.LongTensor)
            and (log_ratio.shape == x.shape[:-1])
        ) or (isinstance(x, torch.LongTensor) and log_ratio.shape == x.shape)
        return log_ratio.mean()

    @abstractmethod
    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    @reset_cache_before_call
    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        batch_loss = self.batch_loss(x=x, y=y)
        assert (
            (not isinstance(x, torch.LongTensor))
            and (batch_loss.shape == x.shape[:-1])
        ) or (isinstance(x, torch.LongTensor) and batch_loss.shape == x.shape)
        return batch_loss.mean()

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Compute an estimation for I(x,y).
        Args:
           x: a tensor with shape [N, D] in which x[i] is sampled from p(x)
           y: a tensor with shape [N, D] in which y[i] is sampled from p(y|x[i])
        Returns:
           estimation: the expected (weighted) log ratio
        """
        return self.mutual_information(x, y)



from functools import partial

import numpy as np
import torch
from pyro.distributions import (
    constraints,
    TransformModule,
    ConditionalTransformModule,
)
from pyro.nn import DenseNN
from torch import nn
from torch.distributions import Transform


class ConditionedLinear(Transform):
    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, params, epsilon=1e-6):
        super(ConditionedLinear, self).__init__(cache_size=1)
        self._params = params
        self._cached_logDetJ = None
        self.epsilon = epsilon

    def _call(self, x):
        loc, log_scale = (
            self._params() if callable(self._params) else self._params
        )
        if not (log_scale.shape == loc.shape):
            log_scale = loc * 0 + log_scale

        self._cached_logDetJ = log_scale.sum(-1)
        y = x * (log_scale.exp() + self.epsilon) + loc
        return y

    def _inverse(self, y):
        loc, log_scale = (
            self._params() if callable(self._params) else self._params
        )
        if not (log_scale.shape == loc.shape):
            log_scale = loc * 0 + log_scale

        x = (y - loc) / (log_scale.exp() + self.epsilon)
        self._cached_logDetJ = log_scale.sum(-1)
        return x

    def log_abs_det_jacobian(self, x, y):
        x_old, y_old = self._cached_x_y
        if x is not x_old or y is not y_old:
            # This call to the parent class Transform will update the cache
            # as well as calling self._call and recalculating y and log_detJ
            self(x)

        return self._cached_logDetJ


class Linear(ConditionedLinear, TransformModule):
    def __init__(
        self, input_dim, loc=None, scale=None, initial_scale=None, epsilon=1e-6
    ):
        super(Linear, self).__init__(self._params)
        if loc is None:
            self.loc = nn.Parameter(torch.zeros(input_dim))
        else:
            self.register_buffer("loc", torch.zeros(input_dim) + loc)
        if initial_scale is None:
            initial_scale = 1.0
        if scale is None:
            self.log_scale = nn.Parameter(
                torch.zeros(input_dim) + np.log(initial_scale)
            )
        else:
            self.register_buffer("log_scale", torch.Tensor([scale]).log())
        self.epsilon = epsilon

    def _params(self):
        return self.loc, self.log_scale


class ConditionalLinear(ConditionalTransformModule):
    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(
        self,
        net,
        loc=None,
        scale=None,
        initial_scale=None,
        epsilon=1e-6,
        skip_connection=False,
    ):
        super(ConditionalLinear, self).__init__()

        self.nn = net
        self.epsilon = epsilon
        self.initial_scale = initial_scale
        self.skip_connection = skip_connection

        if loc is None:
            self.loc = None
        else:
            self.register_buffer("loc", torch.Tensor([loc]).log())

        if initial_scale is None:
            initial_scale = 1.0

        if scale is None:
            self.log_scale = None
            self.initial_scale = initial_scale
        else:
            self.register_buffer("log_scale", torch.Tensor([scale]).log())

    def _params(self, context):
        if self.log_scale is None and self.loc is None:
            loc, log_scale = self.nn(context)
            log_scale = log_scale + np.log(self.initial_scale)
        elif self.loc is None:
            loc = self.nn(context)
            log_scale = self.log_scale
        elif self.log_scale is None:
            loc = self.loc
            log_scale = self.nn(context)
            log_scale = log_scale + np.log(self.initial_scale)
        else:
            loc = self.loc
            log_scale = self.log_scale

        if self.skip_connection and context.shape == loc.shape:
            loc += context

        return loc, log_scale

    def condition(self, context):
        params = partial(self._params, context)
        return ConditionedLinear(params, epsilon=self.epsilon)


def linear(input_dim, loc=None, scale=None, initial_scale=None, **kwargs):
    return Linear(input_dim, scale=scale, initial_scale=initial_scale, loc=loc)


def conditional_linear(
    input_dim,
    context_dim,
    hidden_dims=None,
    scale=None,
    initial_scale=None,
    **kwargs
):
    if hidden_dims is None:
        hidden_dims = [input_dim * 10, input_dim * 10]
    if scale is None:
        nn = DenseNN(
            input_dim=context_dim,
            hidden_dims=hidden_dims,
            param_dims=[input_dim, input_dim],
        )
    else:
        nn = DenseNN(context_dim, hidden_dims, param_dims=[input_dim])
    return ConditionalLinear(nn, scale=scale, initial_scale=initial_scale)



import torch

from torch_mist.baseline import BatchLogMeanExp
from torch_mist.critic import SeparableCritic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class InfoNCE(BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: SeparableCritic,
    ):
        # Note that this can be equivalently obtained by extending TUBA with a BatchLogMeanExp(dim=1) baseline
        # This implementation saves some computation
        super().__init__(
            critic=critic,
            neg_samples=0,  # 0 signifies the whole batch is used as negative samples
            baseline=BatchLogMeanExp("first"),
        )

    def batch_approx_log_partition(
        self, x: torch.Tensor, y: torch.Tensor, f_: torch.tensor
    ):
        # We override the compute_log_normalization just for efficiency
        # The result would be the same as the TUBA implementation with BatchLogMeanExp('first') baseline
        # We override the compute_log_normalization for efficiency since e^(F(x,y))-b(x) = 1
        log_norm = self.baseline(x=x, y=y, f_=f_).unsqueeze(0).expand(f_.shape)
        return log_norm



from typing import Optional, List

import torch
from pyro.nn import DenseNN

from torch import nn

from .base import Critic
from torch_mist.utils.indexing import select_off_diagonal
from torch_mist.utils.shape import expand_to_same_shape


class SeparableCritic(Critic):
    def __init__(
        self,
        f_x: Optional[nn.Module] = None,
        f_y: Optional[nn.Module] = None,
        temperature: float = 1.0,
    ):
        """
        Model the critic as the product of two feature extractors
        f(x,y) = f_x(x)^T f_y(y)
        Note that the baseline can return a constant value (e.g. NWJ estimator)
        :param f_x: a (learnable) model returning a real vector of size D when given x. The identity is used if None is specified.
        :param f_y: a (learnable) model returning a real vector of size D when given y. The identity is used if None is specified.
        """

        super().__init__()
        self.f_x = (lambda x: x) if f_x is None else f_x
        self.f_y = (lambda y: y) if f_y is None else f_y
        self.temperature = temperature

    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        f_x = self.f_x(x)
        f_y = self.f_y(y)

        if f_x.ndim < f_y.ndim:
            f_x = f_x.unsqueeze(0)

        f_x, f_y = expand_to_same_shape(f_x, f_y)

        assert (
            f_x.ndim == f_y.ndim
        ), f"f_x.ndim={f_x.ndim}, f_y.ndim={f_y.ndim}"

        K = torch.einsum("...a, ...a -> ...", f_x, f_y)
        return K / self.temperature



from torch_mist.critic import Critic
from torch_mist.baseline import (
    LearnableBaseline,
    InterpolatedBaseline,
    BatchLogMeanExp,
)
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class AlphaTUBA(BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        baseline: LearnableBaseline,
        alpha: float = 0.5,
        neg_samples: int = -1,
    ):
        alpha_baseline = InterpolatedBaseline(
            baseline_1=BatchLogMeanExp("first"),
            baseline_2=baseline,
            alpha=alpha,
        )

        super().__init__(
            critic=critic,
            baseline=alpha_baseline,
            neg_samples=neg_samples,
        )



from typing import Optional

import torch
from torch.distributions import Distribution
from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.caching import cached, reset_cache_before_call


class DoE(ConditionalGenerativeMIEstimator):
    def __init__(
        self,
        q_Y_given_X: ConditionalDistribution,
        q_Y: Distribution,
    ):
        super().__init__(
            q_Y_given_X=q_Y_given_X,
        )
        self.q_Y = q_Y

    @cached
    def approx_log_p_y(
        self, y: torch.Tensor, x: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        log_q_y = self.q_Y.log_prob(y)
        assert (
            log_q_y.shape == y.shape[:-1]
            and not isinstance(y, torch.LongTensor)
        ) or (log_q_y.shape == y.shape and isinstance(y, torch.LongTensor))
        return log_q_y

    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        log_q_y_given_x = self.approx_log_p_y_given_x(x=x, y=y)
        log_q_y = self.approx_log_p_y(x=x, y=y)

        loss = -log_q_y - log_q_y_given_x

        assert (
            loss.shape == y.shape[:-1] and not isinstance(y, torch.LongTensor)
        ) or (loss.shape == y.shape and isinstance(y, torch.LongTensor))

        return loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  "
            + "(q_Y_given_X): "
            + str(self.q_Y_given_X).replace("\n", "  \n")
            + "\n"
        )
        s += "  " + "(q_Y): " + str(self.q_Y).replace("\n", "  \n") + "\n"
        s += ")" + "\n"
        return s



from typing import List, Any, Callable

from torch import nn

from torch_mist.baseline.base import LearnableBaseline


def baseline_nn(
    x_dim: int, hidden_dims: List[int], nonlinearity: Callable = nn.ReLU(True)
) -> LearnableBaseline:
    from pyro.nn import DenseNN

    net = DenseNN(
        input_dim=x_dim,
        hidden_dims=hidden_dims,
        param_dims=[1],
        nonlinearity=nonlinearity,
    )

    return LearnableBaseline(net)



from typing import Optional, Tuple

import torch


def expand_to_same_shape(*tensors) -> Tuple[torch.Tensor, ...]:
    if len(tensors) == 1:
        assert isinstance(tensors[0], torch.Tensor)
        return tensors[0]
    ndim = tensors[0].ndim

    for tensor in tensors:
        assert tensor.ndim == ndim

    # Find the maximum shape
    max_shape = [
        max([tensor.shape[i] for tensor in tensors]) for i in range(ndim - 1)
    ]
    max_shape += [-1]

    # Expand x and y to the maximum shape
    return tuple([tensor.expand(max_shape) for tensor in tensors])



from typing import Optional

import torch
import torch.nn.functional as F

from torch_mist.baseline import ConstantBaseline
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)
from torch_mist.utils.caching import reset_cache_before_call


class JS(BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
            baseline=ConstantBaseline(value=0.0),
        )

    @reset_cache_before_call
    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> Optional[torch.Tensor]:
        # Compute the critic on the positives. It has shape [...]
        f = self.unnormalized_log_ratio(x=x, y=y)
        # Compute the critic on the negatives. It has shape [M, ...] with M as the number of negative samples
        f_ = self.critic_on_negatives(x=x, y=y)

        loss = F.softplus(-f) + F.softplus(f_).mean(0)
        return loss



from typing import List, Union, Dict

import torch

from torch import nn
from torch.distributions import Transform, Distribution, Normal, Independent
from pyro.distributions import ConditionalTransform

from torch_mist.distributions.joint.wrapper import TorchJointDistribution
from torch_mist.distributions.parametrizations.map import LocScaleMap
from torch_mist.distributions.transforms import (
    TransformedDistributionModule,
    ConditionalTransformedDistributionModule,
    ConditionalDistributionModule,
)


class NormalModule(Distribution, nn.Module):
    def __init__(
        self, loc: torch.Tensor, scale: torch.Tensor, learnable: bool = False
    ):
        assert loc.ndim == 1
        nn.Module.__init__(self)
        Distribution.__init__(
            self, event_shape=torch.Size([loc.shape[0]]), validate_args=False
        )
        if learnable:
            self.loc = nn.Parameter(loc)
            self.log_scale = nn.Parameter(scale.log())
        else:
            self.register_buffer("loc", loc)
            self.register_buffer("log_scale", scale.log())
        self.parametrization = LocScaleMap()

    def rsample(self, sample_shape=torch.Size()):
        params = self.parametrization([self.loc, self.log_scale])
        return Normal(**params).rsample(sample_shape)

    def log_prob(self, value):
        params = self.parametrization([self.loc, self.log_scale])
        return Independent(Normal(**params), 1).log_prob(value)

    def __repr__(self):
        return "Normal()"


class StandardNormalModule(NormalModule):
    def __init__(self, n_dim: int):
        super().__init__(loc=torch.zeros(n_dim), scale=torch.ones(n_dim))


class ConditionalStandardNormalModule(ConditionalDistributionModule):
    def __init__(self, n_dim: int):
        super().__init__()
        self.register_buffer("loc", torch.zeros(n_dim))
        self.register_buffer("log_scale", torch.zeros(n_dim))
        self.parametrization = LocScaleMap()

    def condition(self, context):
        extra_dims = context.ndim - self.loc.ndim
        loc, log_scale = self.loc, self.log_scale
        for _ in range(extra_dims):
            loc = loc.unsqueeze(0)
            log_scale = log_scale.unsqueeze(0)

        loc = loc.expand(*context.shape[:-1], -1)
        log_scale = log_scale.expand(*context.shape[:-1], -1)

        return Independent(Normal(**self.parametrization([loc, log_scale])), 1)


class TransformedNormalModule(TransformedDistributionModule):
    def __init__(self, input_dim: int, transforms: List[Transform]):
        super().__init__(
            base_dist=StandardNormalModule(n_dim=input_dim),
            transforms=transforms,
        )


class ConditionalTransformedNormalModule(
    ConditionalTransformedDistributionModule
):
    def __init__(
        self,
        input_dim: int,
        transforms: List[Union[Transform, ConditionalTransform]],
    ):
        super().__init__(
            base_dist=ConditionalStandardNormalModule(n_dim=input_dim),
            transforms=transforms,
        )


class JointTransformedNormalModule(TorchJointDistribution):
    def __init__(
        self,
        input_dims: Dict[str, int],
        transforms: List[Union[Transform, ConditionalTransform]],
    ):
        splits: List[int] = []
        variables: List[str] = []
        for variable, dim in input_dims.items():
            splits.append(dim)
            variables.append(variable)

        super().__init__(
            torch_dist=TransformedNormalModule(
                input_dim=sum(splits),
                transforms=transforms,
            ),
            variables=variables,
            splits=splits,
            split_dim=-1,
        )



from typing import Dict, Any

import pandas as pd

from torch_mist.utils.logging.logger.base import Logger


class PandasLogger(Logger):
    def __init__(self):
        super().__init__()
        self.__log = []

    def _log(self, data: Dict[str, Any]):
        self.__log.append(data)

    def get_log(self) -> pd.DataFrame:
        return pd.DataFrame(self.__log)

    def _reset(self):
        self.__log = []



from .base import (
    Baseline,
    BatchLogMeanExp,
    LearnableBaseline,
    ConstantBaseline,
    InterpolatedBaseline,
    AlphaTUBABaseline,
    ExponentialMovingAverage,
)
from .factories import baseline_nn



import pyro.distributions.transforms as transforms
import torch
from pyro.distributions import TransformModule

from torch import nn


class Permute(transforms.Permute, TransformModule):
    def __init__(self, permutation, *, dim=- 1, cache_size=1):
        nn.Module.__init__(self)
        transforms.Permute.__init__(self, permutation=permutation,  dim=dim, cache_size=cache_size)
        self.register_buffer('placeholder', torch.zeros(1))

    def update_device(self):
        if self.permutation.device != self.placeholder.device:
            self.permutation = self.permutation.to(self.placeholder.device)

    def _call(self, x):
        self.update_device()
        return super(Permute, self)._call(x)

    def _inverse(self, y):
        self.update_device()
        return super(Permute, self)._inverse(y)

    def log_abs_det_jacobian(self, x, y):
        self.update_device()
        return super(Permute, self).log_abs_det_jacobian(x, y)


def permute(input_dim, permutation=None, dim=-1):
    """
    A helper function to create a :class:`~pyro.distributions.transforms.Permute`
    object for consistency with other helpers.

    :param input_dim: Dimension(s) of input variable to permute. Note that when
        `dim < -1` this must be a tuple corresponding to the event shape.
    :type input_dim: int
    :param permutation: Torch tensor of integer indices representing permutation.
        Defaults to a random permutation.
    :type permutation: torch.LongTensor
    :param dim: the tensor dimension to permute. This value must be negative and
        defines the event dim as `abs(dim)`.
    :type dim: int

    """
    if dim < -1 or not isinstance(input_dim, int):
        if len(input_dim) != -dim:
            raise ValueError(
                "event shape {} must have same length as event_dim {}".format(
                    input_dim, -dim
                )
            )
        input_dim = input_dim[dim]

    if permutation is None:
        permutation = torch.randperm(input_dim)
    return Permute(permutation, dim=dim)



import torch


def select_off_diagonal(x: torch.Tensor, K: int) -> torch.Tensor:
    N = x.shape[0]

    # Consider the first K off-diagonal elements
    idx = torch.arange(N * K).to(x.device).view(N, K).long()
    idx = (idx % K + torch.div(idx, K, rounding_mode="floor") + 1) % N
    return x[idx.T]



from typing import Optional, List

import torch
from pyro.distributions import ConditionalDistribution
from torch import nn

from torch_mist.quantization.functions import (
    LearnableVectorQuantization,
    QuantizationFunction,
)

INITIAL_PATIENCE = 10.0


class VQVAE(nn.Module):
    def __init__(
        self,
        encoder: LearnableVectorQuantization,
        decoder: ConditionalDistribution,
        beta: float = 0.25,
        gamma: float = 0.99,
        cross_modal: bool = False,
    ):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder

        n_bins = encoder.n_bins

        self.beta = beta
        self.gamma = gamma
        self.cross_modal = cross_modal

        initial_assignments = torch.zeros(n_bins) + INITIAL_PATIENCE
        self.register_buffer("assignments", initial_assignments)
        self.min_assignment = 0.1 / n_bins

    @property
    def quantization(self) -> QuantizationFunction:
        return self.encoder

    def loss(
        self, x: torch.Tensor, y: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        assert (
            not self.cross_modal or y is not None
        ), "y must be provided if cross_modal is True"

        z = self.encoder.net(x)
        indices = self.encoder.codebook_lookup(z)

        quantized = self.encoder.vectors[indices]
        z_q = z + (quantized - z).detach()

        assert (
            z_q.shape == z.shape == quantized.shape
        ), f"{z_q.shape} != {z.shape} != {quantized.shape}"

        if self.cross_modal:
            reconstruction_loss = -self.decoder.condition(z_q).log_prob(y)
            assert (
                reconstruction_loss.shape == y.shape[:-1]
            ), f"{reconstruction_loss.shape} != {y.shape[:-1]}"
            reconstruction_loss = reconstruction_loss.mean()
        else:
            reconstruction_loss = -self.decoder.condition(z_q).log_prob(x)
            assert (
                reconstruction_loss.shape == x.shape[:-1]
            ), f"{reconstruction_loss.shape} != {x.shape[:-1]}"
            reconstruction_loss = reconstruction_loss.mean()

        if self.training:
            # Compute the one-hot representation of the indices [N*M, n_bins]
            flat_indices = indices.view(-1)
            one_hot = torch.nn.functional.one_hot(
                flat_indices, self.encoder.n_bins
            ).float()

            # Compute the number of datapoints assigned to each bin [n_bins]
            assignments = one_hot.mean(0)

            if self.training:
                self.assignments = (
                    self.gamma * self.assignments
                    + (1 - self.gamma) * assignments
                )
            # assignments = self.assignments

            # Compute the mean vector for each assignment [n_bins, z_dim]
            N = flat_indices.shape[0]
            mean_vector = (one_hot.unsqueeze(-1) * z.unsqueeze(-2)).sum(0)
            mean_vector = mean_vector / (assignments.unsqueeze(-1) * N)

            not_assigned = [
                i.item()
                for i in torch.arange(self.encoder.n_bins)[
                    self.assignments * N < self.min_assignment
                ].long()
            ]
            while len(not_assigned) > 0:
                not_assigned_idx = not_assigned.pop()
                # Select the vector with the highest assignment
                max_idx = torch.argmax(assignments)
                # set the centroid to the vector with the highest assignment + some noise
                noise = torch.randn_like(self.encoder.vectors[max_idx]) * 1e-5
                self.encoder.vectors[not_assigned_idx] = (
                    self.encoder.vectors[max_idx] + noise
                )
                # set the new assignements to half the max assignment
                new_assignment = self.assignments[max_idx] / 2
                self.assignments[not_assigned_idx] = new_assignment
                self.assignments[max_idx] = new_assignment

            mean_vector[assignments == 0] = self.encoder.vectors[
                assignments == 0
            ]

            self.encoder.vectors = (
                self.gamma * self.encoder.vectors
                + (1 - self.gamma) * mean_vector
            )

        commitment_loss = torch.mean((quantized.detach() - z) ** 2)
        # codebook_loss = torch.mean((quantized - z.detach())**2)
        loss = (
            reconstruction_loss + self.beta * commitment_loss
        )  # + codebook_loss
        return loss



from typing import List, Iterator, Optional, Dict, Any, Union

import numpy as np
import torch
from sklearn.cluster import KMeans

from torch_mist.distributions.factories import conditional_transformed_normal
from torch_mist.quantization.functions import LearnableVectorQuantization
from torch_mist.quantization.functions import ClusterQuantization
from torch_mist.quantization.vqvae import VQVAE


def vector_quantization(
    input_dim: int,
    n_bins: int,
    hidden_dims: List[int],
    quantization_dim: Optional[int] = None,
) -> LearnableVectorQuantization:
    assert len(hidden_dims) > 0, "hidden_dims must be a non-empty list"

    from pyro.nn import DenseNN

    if quantization_dim is None:
        quantization_dim = 16

    encoder = DenseNN(input_dim, hidden_dims, [quantization_dim])

    quantization = LearnableVectorQuantization(
        net=encoder, n_bins=n_bins, quantization_dim=quantization_dim
    )

    return quantization


def kmeans_quantization(
    data: Union[torch.Tensor, np.ndarray], n_bins: int, **kwargs
) -> ClusterQuantization:
    clustering = KMeans(n_bins, **kwargs)
    clustering.fit(data)

    return ClusterQuantization(clustering)


def vqvae(
    x_dim: int,
    quantization_dim: int,
    n_bins: int,
    hidden_dims: List[int],
    y_dim: Optional[int] = None,
    cross_modal: bool = False,
    beta: float = 0.2,
) -> VQVAE:
    assert (
        not cross_modal or y_dim is not None
    ), "If cross_modal is True, y_dim must be provided"

    quantization = vector_quantization(
        input_dim=x_dim,
        quantization_dim=quantization_dim,
        hidden_dims=hidden_dims,
        n_bins=n_bins,
    )

    decoder = conditional_transformed_normal(
        input_dim=x_dim if not cross_modal else y_dim,
        context_dim=quantization_dim,
        transform_name="conditional_linear",
        hidden_dims=hidden_dims,
    )

    return VQVAE(
        encoder=quantization,
        decoder=decoder,
        cross_modal=cross_modal,
        beta=beta,
    )


def vqvae_quantization(
    input_dim: int,
    n_bins: int,
    hidden_dims: List[int],
    data: Optional[Union[torch.Tensor, np.array]] = None,
    dataloader: Optional[Iterator] = None,
    quantization_dim: Optional[int] = None,
    beta: float = 0.2,
    max_epochs: int = 1,
    optimizer_class=torch.optim.Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> LearnableVectorQuantization:
    from torch_mist.utils.train.model import train_model

    if optimizer_params is None:
        optimizer_params = {"lr": 1e-3}

    quantization = vector_quantization(
        input_dim=input_dim,
        n_bins=n_bins,
        hidden_dims=hidden_dims,
        quantization_dim=quantization_dim,
    )

    decoder = conditional_transformed_normal(
        input_dim=input_dim,
        context_dim=quantization_dim,
        transform_name="conditional_linear",
        hidden_dims=hidden_dims,
        n_transforms=1,
    )

    model = VQVAE(encoder=quantization, decoder=decoder, beta=beta)

    train_model(
        model=model,
        dataloader=dataloader,
        data=data,
        max_epochs=max_epochs,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        batch_size=batch_size,
        num_workers=num_workers,
        logger=False,
    )

    return quantization



from .base import (
    DistributionModule,
    ConditionalDistributionModule,
    TransformedDistributionModule,
    ConditionalTransformedDistributionModule,
)
from .linear import Linear, ConditionalLinear, linear, conditional_linear
from .permute import Permute, permute
from .utils import fetch_transform



# from .base import ParametrizedDistribution, ConstantParametrizedConditionalDistribution
# from .bernoulli import ParametrizedBernoulli
# from .delta import ParametrizedDelta
# from .normal import ParametrizedNormal





from typing import Dict, Union
from collections.abc import Iterator

import numpy as np
import torch
from torch.distributions import Distribution
from torch.utils.data import DataLoader, Dataset

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.utils.data.sampler import SameAttributeSampler


class DistributionDataLoader(Iterator[Dict[str, torch.Tensor]]):
    def __init__(
        self,
        joint_dist: Union[Distribution, JointDistribution],
        batch_size: int,
        max_samples: int = 100000,
        split_dim: int = -1,
    ):
        self.joint_dist = joint_dist
        self.split_dim = split_dim
        self.batch_size = batch_size
        self.max_samples = max_samples
        self._n_samples = max_samples

    def __next__(self) -> Dict[str, torch.Tensor]:
        if self._n_samples <= 0:
            self._n_samples = self.max_samples
            raise StopIteration
        else:
            n_samples = min(self.batch_size, self._n_samples)

            samples = self.joint_dist.sample(torch.Size([n_samples]))

            if isinstance(samples, tuple):
                assert len(samples) == 2
                samples = {"x": samples[0], "y": samples[1]}

            self._n_samples -= n_samples

            assert isinstance(samples, dict)

            return samples

    def __iter__(self):
        return DistributionDataLoader(
            joint_dist=self.joint_dist,
            batch_size=self.batch_size,
            max_samples=self.max_samples,
            split_dim=self.split_dim,
        )

    def __len__(self):
        return self.max_samples // self.batch_size


class SameAttributeDataLoader(DataLoader):
    def __init__(
        self,
        dataset: Dataset,
        attributes: Union[torch.Tensor, np.ndarray],
        batch_size: int,
        **kwargs
    ):
        if len(dataset) != len(attributes):
            raise Exception(
                "The dataset and the attributes must have the same length."
            )

        super().__init__(
            dataset,
            batch_sampler=SameAttributeSampler(
                attributes=attributes, batch_size=batch_size
            ),
            **kwargs
        )


def sample_same_attributes(
    dataloader: DataLoader,
    attributes: Union[torch.Tensor, np.ndarray],
) -> DataLoader:
    return SameAttributeDataLoader(
        dataset=dataloader.dataset,
        attributes=attributes,
        batch_size=dataloader.batch_size,
        num_workers=dataloader.num_workers,
        prefetch_factor=dataloader.prefetch_factor,
        pin_memory=dataloader.pin_memory,
        pin_memory_device=dataloader.pin_memory_device,
        timeout=dataloader.timeout,
        worker_init_fn=dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
    )



from .factories import *
from .normal import (
    NormalModule,
    StandardNormalModule,
    TransformedNormalModule,
    ConditionalTransformedNormalModule,
    JointTransformedNormalModule,
)
from .categorical import CategoricalModule, ConditionalCategoricalModule



from typing import Optional, Callable, Any, Dict, Union, List

import torch
from torch import nn

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.caching import (
    reset_cache_before_call,
    reset_cache_after_call,
    cached,
)
from torch_mist.utils.freeze import is_frozen


class DummyModule(nn.Module):
    def __init__(self, f: Callable[[Any], Any]):
        super().__init__()
        self.f = f

    def forward(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def __repr__(self):
        return str(self.f)


class TransformedMIEstimator(MIEstimator):
    def __init__(
        self,
        base_estimator: MIEstimator,
        transforms: Dict[str, Callable[[Any], Any]],
    ):
        super().__init__()

        self.infomax_gradient = base_estimator.infomax_gradient

        self.base_estimator = base_estimator
        self.transforms = nn.ModuleDict()

        for variable, transform in transforms.items():
            if not base_estimator.infomax_gradient and (
                not is_frozen(transform)
            ):
                raise ValueError(
                    "Transforms can be trained together with the estimator only when the estimator provides a valid infomax gradient."
                    f"The estimator {base_estimator.__class__} does not. You can use a different estimator or the transform form {variable}."
                )

            if not isinstance(transform, nn.Module):
                transform = DummyModule(transform)

            self.transforms[variable] = transform

    @cached
    def transform(self, **variables) -> Dict[str, torch.Tensor]:
        transformed_variables = {}
        for variable, transform in self.transforms.items():
            transformed_variables[variable] = transform(variables[variable])
        return transformed_variables

    def batch_loss(self, *args, **variables) -> torch.Tensor:
        if len(args) != 0 and len(args) != 2:
            raise ValueError(
                "The loss method can be called by passing two arguments or by specifying multiple named named arguments."
            )
        if len(args) == 2:
            if len(variables) > 0:
                raise ValueError(
                    "The loss method can be called by passing two arguments or by specifying multiple named named arguments."
                )
            variables["x"] = args[0]
            variables["y"] = args[1]

        variables.update(self.transform(**variables))
        return self.base_estimator.batch_loss(**variables)

    def log_ratio(
        self, *args, **variables
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        if len(args) != 0 and len(args) != 2:
            raise ValueError(
                "The log_ratio method can be called by passing two arguments or by specifying multiple named named arguments."
            )
        if len(args) == 2:
            if len(variables) > 0:
                raise ValueError(
                    "The log_ratio method can be called by passing two arguments or by specifying multiple named named arguments."
                )
            variables["x"] = args[0]
            variables["y"] = args[1]

        variables.update(self.transform(**variables))
        return self.base_estimator.log_ratio(**variables)

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self.mutual_information(*args, **kwargs)



import math
import torch

from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.caching import (
    reset_cache_after_call,
    reset_cache_before_call,
)
from torch_mist.utils.indexing import select_off_diagonal


class L1Out(ConditionalGenerativeMIEstimator):
    def __init__(
        self, q_Y_given_X: ConditionalDistribution, neg_samples: int = -1
    ):
        super().__init__(q_Y_given_X=q_Y_given_X)
        self.neg_samples = neg_samples

    def _broadcast_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        assert (
            x.shape[0] == y.shape[0]
        ), "The batch size of x and y must be the same"
        N = x.shape[0]

        neg_samples = self.neg_samples
        if neg_samples <= 0:
            neg_samples = N + neg_samples

        if neg_samples < 1:
            neg_samples = 1

        if self.neg_samples == 0 or self.neg_samples == -1:
            x = x.unsqueeze(0)
        else:
            x = select_off_diagonal(x, neg_samples)

        # Probability of all the other y in the same batch [M, N, ...]
        p_y_given_X = self.q_Y_given_x(x=x)
        log_p_y_given_x = p_y_given_X.log_prob(y.unsqueeze(1))

        if self.neg_samples == -1:
            D = torch.eye(N).to(y.device)
            while D.ndim < log_p_y_given_x.ndim:
                D = D.unsqueeze(-1)

            # Remove the diagonal
            log_p_y_given_x = log_p_y_given_x * (1 - D)

            # Set the diagonal to -inf
            log_p_y_given_x = log_p_y_given_x + torch.nan_to_num(
                D * (-float("inf")),
                0,
                float("inf"),
                -float("inf"),
            )

        return log_p_y_given_x

    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # log prob of the samples in the batch, shape [M, N, ...]
        log_p_y_given_x = self._broadcast_log_p_y_given_x(x, y)

        # Compute the expectation using logsumexp. The shape is [N, ...]
        log_p_y = torch.logsumexp(log_p_y_given_x, dim=0) - math.log(
            log_p_y_given_x.shape[0]
        )

        assert log_p_y.shape == y.shape[:-1]

        return log_p_y

    def log_ratio(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        log_p_y_x = self.approx_log_p_y_given_x(x=x, y=y)
        log_p_y = self.approx_log_p_y(x=x, y=y)

        assert (
            log_p_y_x.ndim == log_p_y.ndim
        ), f"log_p_y_x.ndim={log_p_y_x.ndim}, log_p_y.ndim={log_p_y.ndim}"
        log_ratio = log_p_y_x - log_p_y

        return log_ratio

    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        log_q_y_given_x = self.approx_log_p_y_given_x(x=x, y=y)
        loss = -log_q_y_given_x

        assert (
            loss.shape == y.shape[:-1] and isinstance(y, torch.FloatTensor)
        ) or (loss.shape == y.shape and isinstance(y, torch.LongTensor))

        return loss



import inspect
from typing import List, Union, Dict

from pyro.distributions import ConditionalTransform, ConditionalDistribution
from pyro.nn import DenseNN

from torch.distributions import Transform, Distribution

from torch_mist.distributions.categorical import ConditionalCategoricalModule
from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.distributions.normal import (
    ConditionalTransformedNormalModule,
    TransformedNormalModule,
    JointTransformedNormalModule,
)

from torch_mist.distributions.transforms.utils import fetch_transform


def make_transforms(
    input_dim: int,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
    **kwargs,
) -> List[Union[Transform, ConditionalTransform]]:
    assert n_transforms > 0, "n_transforms must be greater than 0"
    transforms = []

    transform_factory = fetch_transform(transform_name)

    # Check the arguments
    kwargs_to_delete = []
    for arg_name in kwargs:
        if not (arg_name in inspect.signature(transform_factory).parameters):
            print(
                f"Warning: parameter {arg_name} ignored for {transform_name}."
            )
            kwargs_to_delete.append(arg_name)

    for arg_name in kwargs_to_delete:
        del kwargs[arg_name]

    for transform in range(n_transforms):
        transform = transform_factory(input_dim=input_dim, **kwargs)
        transforms.append(transform)

    return transforms


def transformed_normal(
    input_dim: int,
    transform_name: str = "linear",
    n_transforms: int = 1,
    **kwargs,
) -> Distribution:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    transforms = make_transforms(
        input_dim=input_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        **kwargs,
    )

    return TransformedNormalModule(
        input_dim=input_dim,
        transforms=transforms,
    )


def conditional_transformed_normal(
    input_dim: int,
    context_dim: int,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
    **kwargs,
) -> ConditionalDistribution:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    transforms = make_transforms(
        input_dim=input_dim,
        context_dim=context_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        **kwargs,
    )

    return ConditionalTransformedNormalModule(
        input_dim=input_dim, transforms=transforms
    )


def joint_transformed_normal(
    input_dims: Dict[str, int],
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
    **kwargs,
) -> JointDistribution:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    input_dim = sum(input_dims.values())

    transforms = make_transforms(
        input_dim=input_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        **kwargs,
    )

    return JointTransformedNormalModule(
        input_dims=input_dims, transforms=transforms
    )


def conditional_categorical(
    n_classes: int,
    context_dim: int,
    hidden_dims: List[int],
    temperature: float = 1.0,
):
    net = DenseNN(
        input_dim=context_dim, hidden_dims=hidden_dims, param_dims=[n_classes]
    )
    return ConditionalCategoricalModule(net, temperature=temperature)



from typing import Any

from torch import nn


def freeze(function: Any) -> Any:
    if isinstance(function, nn.Module):
        for param in function.parameters():
            param.requires_grad = False

    return function


def n_trainable_parameters(function: Any) -> int:
    if isinstance(function, nn.Module):
        n_parameters = 0
        for param in function.parameters():
            if param.requires_grad:
                n_parameters += param.numel()

        return n_parameters
    else:
        return 0


def is_frozen(function: Any) -> bool:
    if isinstance(function, nn.Module):
        return n_trainable_parameters(function) == 0
    else:
        return True



import torch
from torch.distributions import Distribution
from torch_mist.utils.indexing import select_off_diagonal


class EmpiricalDistribution(Distribution):
    def __init__(self):
        super().__init__(validate_args=False)
        self._samples = None

    def add_samples(self, samples):
        self._samples = samples

    def sample(self, sample_shape=torch.Size()):
        assert self._samples is not None
        assert len(sample_shape) == 1
        n_samples = sample_shape[0]

        return select_off_diagonal(self._samples, n_samples)

    def update(self):
        self._samples = None



from torch_mist.baseline import BatchLogMeanExp, ExponentialMovingAverage
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
    CombinedDiscriminativeMIEstimator,
)


class MINE(CombinedDiscriminativeMIEstimator):
    lower_bound = False  # Technically MINE is a lower bound, but sometimes it converges from above

    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
        gamma: float = 0.9,
    ):
        super().__init__(
            train_estimator=BaselineDiscriminativeMIEstimator(
                critic=critic,
                neg_samples=neg_samples,
                baseline=ExponentialMovingAverage(gamma=gamma),
            ),
            eval_estimator=BaselineDiscriminativeMIEstimator(
                critic=critic,
                neg_samples=neg_samples,
                baseline=BatchLogMeanExp("all"),
            ),
        )



from typing import TypeVar, Optional, Union

from typing import List, Dict

import torch
from pyro.distributions import ConditionalDistribution
from torch.distributions import Distribution

from torch import nn

T = TypeVar("T", bound="NamedDistribution")


class JointDistribution(nn.Module, Distribution, ConditionalDistribution):
    def __init__(
        self,
        variables: List[str],
        name: str = "p",
    ):
        nn.Module.__init__(self)
        Distribution.__init__(self, validate_args=False)

        self.variables = variables
        self.name = name

    def __repr__(self):
        return f"{self.name}({','.join(self.variables)})"

    def _log_prob(self, **kwargs) -> torch.Tensor:
        raise NotImplementedError()

    def log_prob(self, *args, **kwargs) -> torch.Tensor:
        AMBIGUOUS_CALL_ERROR = "Please specify the variable names to avoid ambiguities. E.g. p.log_prob(x=values_x, y=values_y)"

        if len(args) > 1:
            raise ValueError(AMBIGUOUS_CALL_ERROR)
        elif len(args) == 1:
            if len(self.variables) > 1:
                raise ValueError(AMBIGUOUS_CALL_ERROR)
            elif self.variables[0] in kwargs:
                raise ValueError(AMBIGUOUS_CALL_ERROR)
            kwargs[self.variables[0]] = args[0]

        return self._log_prob(**kwargs)

    def _marginal(self: T, variables: List[str]) -> T:
        raise NotImplementedError()

    def marginal(self: T, *variables) -> T:
        valid_labels: List[str] = []
        for label in variables:
            if not (label in self.variables):
                raise ValueError(
                    f"Can not compute the marginal p({label}) since the distribution is defined on p({','.join(self.variables)})"
                )
            valid_labels.append(label)
        return self._marginal(valid_labels)

    def conditional(self, *variables) -> ConditionalDistribution:
        valid_variables: List[str] = []
        for variable in variables:
            if variable not in self.variables:
                raise ValueError(
                    f"The variable {variable} is not part of the support of {self.__repr__()}."
                )
            valid_variables.append(variable)
        return ConditionalNamedDistribution(self, valid_variables)

    def condition(self, **conditioning) -> T:
        valid_variables: List[str] = []
        for variable in conditioning:
            if variable not in self.variables:
                raise ValueError(
                    f"The variable {variable} is not part of the support of {self.__repr__()}."
                )
            valid_variables.append(variable)
        return ConditionedJointDistribution(
            joint=self,
            marginal=self.marginal(*valid_variables),
            cond_dict=conditioning,
        )

    def _mutual_information(
        self, variable_1: str, variable_2: str
    ) -> torch.Tensor:
        mi = (
            self.entropy(variable_1)
            + self.entropy(variable_2)
            - self.entropy(variable_1, variable_2)
        )
        return mi

    def mutual_information(
        self,
        variable_1: Optional[str] = None,
        variable_2: Optional[str] = None,
    ) -> torch.Tensor:
        if variable_1 is None:
            if not (variable_2) is None:
                raise ValueError(
                    "Please both variable for mutual information computation."
                )
            if not len(self.variables) == 2:
                raise ValueError(
                    "Please which variable should be used for mutual information computation."
                )
            variable_1 = self.variables[0]
            variable_2 = self.variables[1]

        if not (variable_1 in self.variables) or not (
            variable_2 in self.variables
        ):
            raise ValueError(
                f"Can not compute I({variable_1};{variable_2}) since the distribution is defined on p({','.join(self.variables)})"
            )
        return self._mutual_information(variable_1, variable_2)

    def _entropy(self, variables: List[str]) -> torch.Tensor:
        raise NotImplementedError()

    def entropy(self, *variables) -> torch.Tensor:
        valid_labels: List[str] = []
        for variable in variables:
            assert isinstance(variable, str)
            if not (variable in self.variables):
                raise ValueError(
                    f"Can not compute H({variables})since the distribution is defined on p({','.join(self.variables)})"
                )
            valid_labels.append(variable)
        return self._entropy(valid_labels)


class ConditionedJointDistribution(JointDistribution):
    def __init__(
        self,
        joint: JointDistribution,
        marginal: JointDistribution,
        cond_dict: Dict[str, torch.Tensor],
    ):
        if joint.name != marginal.name:
            raise ValueError("The two distribution must have the same name")

        super().__init__(
            name=joint.name,
            variables=[
                variable
                for variable in joint.variables
                if not (variable in cond_dict)
            ],
        )
        self.joint = joint
        self.log_marginal = marginal.log_prob(**cond_dict)
        self.cond_dict = cond_dict

    def _log_prob(self, **kwargs) -> torch.Tensor:
        kwargs = {**self.cond_dict, **kwargs}
        log_joint = self.joint.log_prob(**kwargs)
        log_conditional = log_joint - self.log_marginal
        return log_conditional

    def __repr__(self):
        s = super().__repr__()[:-1]
        s += f"|{','.join(self.cond_dict)})"
        return s


class ConditionalNamedDistribution(ConditionalDistribution):
    def __init__(
        self,
        joint: JointDistribution,
        condition_on: List[str],
    ):
        super().__init__()
        self.joint = joint
        self.condition_on = condition_on

    def condition(
        self, context: Union[Dict[str, torch.Tensor], torch.Tensor]
    ) -> JointDistribution:
        if isinstance(context, torch.Tensor):
            assert len(self.condition_on) == 1
            context = {self.condition_on[0]: context}
        else:
            assert isinstance(context, dict)

        return ConditionedJointDistribution(
            joint=self.joint,
            marginal=self.joint.marginal(*self.condition_on),
            cond_dict=context,
        )

    def __repr__(self):
        s = f"{self.joint.name}({','.join([variable for variable in self.joint.variables if not (variable in self.condition_on)])}|"
        s += f"{','.join(self.condition_on)})"
        return s



import torch

from torch_mist.estimators import MIEstimator


class FlippedMIEstimator(MIEstimator):
    def __init__(self, estimator: MIEstimator):
        super().__init__()
        self.estimator = estimator

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.estimator.log_ratio(y, x)

    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.estimator.loss(y, x)

    def mutual_information(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        return self.estimator.mutual_information(y, x)


def flip_estimator(estimator: MIEstimator) -> MIEstimator:
    if isinstance(estimator, FlippedMIEstimator):
        return estimator.estimator
    else:
        return FlippedMIEstimator(estimator)



from abc import abstractmethod
from typing import Optional, Union

import torch
from pyro.distributions import ConditionalDistribution
from torch.distributions import Distribution

from torch_mist.baseline import Baseline
from torch_mist.distributions.empirical import EmpiricalDistribution
from torch_mist.estimators.base import MIEstimator
from torch_mist.critic import Critic
from torch_mist.critic import SeparableCritic
from torch_mist.utils.caching import (
    cached,
    reset_cache_before_call,
    reset_cache_after_call,
)
from torch_mist.utils.indexing import select_off_diagonal


class DiscriminativeMIEstimator(MIEstimator):
    lower_bound: bool = True
    infomax_gradient: bool = True

    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
        proposal: Optional[
            Union[Distribution, ConditionalDistribution]
        ] = None,
    ):
        super().__init__()
        self.critic = critic
        self.neg_samples = neg_samples
        if proposal is None:
            proposal = EmpiricalDistribution()
        self.proposal = proposal

    @cached
    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        f = self.critic(x, y)
        assert f.ndim == y.ndim - 1
        return f

    def get_n_negatives(self, N: int):
        neg_samples = self.neg_samples

        # Negative neg_samples values are interpreted as difference from the batch size (-1 is all but one in the batch)
        if neg_samples <= 0:
            neg_samples = N + neg_samples

        # We can't use more negative than the batch
        if neg_samples > N:
            neg_samples = N

        # At least one negative sample
        neg_samples = max(neg_samples, 1)
        return neg_samples

    @cached
    def critic_on_negatives(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        if isinstance(self.proposal, EmpiricalDistribution):
            assert y is not None
            self.proposal.add_samples(y)

        N = x.shape[0]
        neg_samples = self.get_n_negatives(N)

        # If we are sampling using the empirical distribution and the critic is separable
        # We can use an efficient implementation
        if isinstance(self.proposal, Distribution) and isinstance(
            self.critic, SeparableCritic
        ):
            # Efficient implementation for separable critic with empirical distribution (negatives from the same batch)
            if isinstance(self.proposal, EmpiricalDistribution):
                # Keep only neg_samples negatives for each positive (off-diagonal)
                if neg_samples < N:
                    # Override the default self.critic(x,y) for efficient computation
                    f_x = self.critic.f_x(x)
                    f_y = self.critic.f_y(y)
                    off_diagonal_f_y = select_off_diagonal(f_y, neg_samples)
                    f_ = torch.einsum(
                        "ab...c, b...c -> ab...", off_diagonal_f_y, f_x
                    )
                else:
                    # Take the M stored samples from y (from the empirical). Shape [M, 1, ...,Y_DIM]
                    y_ = self.proposal._samples[:neg_samples].unsqueeze(1)
                    # Compute the critic on all pairs
                    f_ = self.critic(x, y_)
            else:
                # Efficient implementation for separable critic with unconditional proposal
                # Here we re-use the same samples y'~ r(y) for all the batch
                # Sample from the proposal r(y) [M', 1, ..., Y_DIM] with M' as the number of neg_samples
                y_ = self.proposal.sample(
                    sample_shape=torch.Size([neg_samples])
                ).unsqueeze(1)

                # Compute the critic on them. Shape [M', ...]
                f_ = self.critic(x, y_)
        else:
            # Sample from the proposal r(y|x) [M, ..., Y_DIM] with M as the number of neg_samples
            if isinstance(self.proposal, ConditionalDistribution):
                proposal = self.proposal.condition(x)
            else:
                proposal = self.proposal

            y_ = proposal.sample(sample_shape=torch.Size([neg_samples]))
            # The shape of the samples from the proposal distribution is [M, ..., Y_DIM]
            assert y_.ndim == x.ndim + 1 and y_.shape[0] == neg_samples
            assert y_.shape[0] == neg_samples and y_.ndim == x.ndim + 1

            # Compute the log-ratio on the samples from the proposal p(x)r(y|x) [M,...]
            f_ = self.critic(x, y_)

        assert (
            f_.shape[0] == neg_samples
        ), f"Expected {neg_samples} samples, got {f_.shape[0]}"
        assert (
            f_.shape[1:] == x.shape[:-1]
        ), f"Negatives have shape {f_.shape} shape, while x has shape {x.shape}"

        if isinstance(self.proposal, EmpiricalDistribution):
            self.proposal.update()

        return f_

    def approx_log_partition(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        # Evaluate the unnormalized_log_ratio f(x,y) on the samples from p(x)r(y|x)
        # The tensor f_ has shape [M, N...] in which f_[i,j] contains critic(x[j], y_[i,j]).
        # and y_ is sampled from r(y|x), which is set to the empirical p(y) unless a proposal is specified
        f_ = self.critic_on_negatives(x, y)

        log_Z = self.batch_approx_log_partition(x, y, f_)

        N = x.shape[0]
        assert log_Z.shape[0] == self.get_n_negatives(N)
        assert (
            not isinstance(x, torch.LongTensor)
            and log_Z.shape[1:] == x.shape[:-1]
        ) or (isinstance(x, torch.LongTensor) and log_Z.shape[1:] == x.shape)

        return log_Z.mean(0)

    @abstractmethod
    def batch_approx_log_partition(
        self, x: torch.Tensor, y: torch.Tensor, f_: torch.Tensor
    ):
        raise NotImplementedError()

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        assert x.ndim == y.ndim
        # Approximate the log-ratio p(y|x)/p(y) on samples from p(x,y).
        # x and y have shape [..., X_DIM] and [..., Y_DIM] respectively

        # Evaluate the unnormalized_log_ratio f(x,y) on the samples from p(x, y), with shape [...]
        unnormalized_log_ratio = self.unnormalized_log_ratio(x, y)

        # Compute the log-normalization term, with shape [M, ...]
        log_partition = self.approx_log_partition(x, y)

        log_ratio = unnormalized_log_ratio - log_partition
        assert log_ratio.ndim == y.ndim - 1

        return log_ratio

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -self.log_ratio(x, y)

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (critic): " + str(self.critic).replace("\n", "\n" + "  ") + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s


class CombinedDiscriminativeMIEstimator(DiscriminativeMIEstimator):
    def __init__(
        self,
        train_estimator: DiscriminativeMIEstimator,
        eval_estimator: DiscriminativeMIEstimator,
    ):
        assert train_estimator.critic == eval_estimator.critic
        assert train_estimator.neg_samples == eval_estimator.neg_samples

        super().__init__(
            critic=train_estimator.critic,
            neg_samples=train_estimator.neg_samples,
        )

        self.train_estimator = train_estimator
        self.eval_estimator = eval_estimator

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.train_estimator.batch_loss(x, y)

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.eval_estimator.log_ratio(x, y)


class BaselineDiscriminativeMIEstimator(DiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        baseline: Baseline,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
        )

        self.baseline = baseline

    def batch_approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
    ) -> torch.Tensor:
        # Compute the baseline. It has shape [1,...]
        b = self.baseline(f_, x, y).unsqueeze(0)
        assert (
            b.ndim == f_.ndim
        ), f"Baseline has ndim {b.ndim} while f_ has ndim {f_.ndim}"

        log_norm = (f_ - b).exp() + b - 1.0

        return log_norm

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (ratio_estimator): "
            + str(self.critic).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += (
            "  (baseline): "
            + str(self.baseline).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s



from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)
from torch_mist.critic import Critic
from torch_mist.baseline import ConstantBaseline


class NWJ(BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
            baseline=ConstantBaseline(value=1.0),
        )



from typing import Optional, List

from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.transformed.implementations import (
    PQ,
    BinnedMIEstimator,
)
from torch_mist.quantization import QuantizationFunction


def binned(
    Q_x: Optional[QuantizationFunction] = None,
    Q_y: Optional[QuantizationFunction] = None,
    temperature: float = 0.1,
) -> BinnedMIEstimator:
    return BinnedMIEstimator(
        Q_x=Q_x,
        Q_y=Q_y,
        temperature=temperature,
    )


def pq(
    Q_y: QuantizationFunction,
    x_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_QY_given_X: Optional[ConditionalDistribution] = None,
    temperature: float = 0.1,
) -> PQ:
    from torch_mist.distributions.factories import conditional_categorical

    if q_QY_given_X is None:
        if x_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_QY_given_X or x_dim and hidden_dims must be specified."
            )
        q_QY_given_X = conditional_categorical(
            n_classes=Q_y.n_bins,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            temperature=temperature,
        )

    return PQ(
        q_QY_given_X=q_QY_given_X,
        Q_y=Q_y,
        temperature=temperature,
    )



from typing import Optional

from sklearn.base import TransformerMixin
import torch
from torch import nn


class QuantizationFunction(nn.Module):
    @property
    def n_bins(self) -> int:
        raise NotImplementedError()

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        raise NotImplementedError()


class LearnableQuantization(QuantizationFunction):
    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()


class ClusterQuantization(QuantizationFunction):
    def __init__(self, clustering: TransformerMixin):
        super().__init__()
        self.clustering = clustering

    @property
    def n_bins(self) -> int:
        return self.clustering.n_clusters

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        device = x.device
        if isinstance(x, torch.Tensor):
            x = x.cpu()

        shape = x.shape[:-1]
        feature_dim = x.shape[-1]

        return (
            torch.LongTensor(self.clustering.predict(x.view(-1, feature_dim)))
            .view(shape)
            .to(device)
        )


class VectorQuantization(QuantizationFunction):
    def __init__(
        self,
        input_dim: Optional[int] = None,
        n_bins: Optional[int] = None,
        vectors: Optional[torch.Tensor] = None,
    ):
        super().__init__()

        if vectors is None:
            if n_bins is None or input_dim is None:
                raise ValueError(
                    "Either `vectors` or `n_bins` and `input_dim` need to be specified"
                )

            # Vectors used for quantization
            vectors = torch.zeros(n_bins, input_dim)
            vectors.uniform_(-1 / n_bins, 1 / n_bins)
            # self.vectors = nn.Parameter(vectors)
        else:
            if not (n_bins is None or input_dim is None):
                raise ValueError(
                    "Either `vectors` or `n_bins` and `input_dim` need to be specified"
                )
        self.register_buffer("vectors", vectors)

    def codebook_lookup(self, x: torch.Tensor) -> torch.Tensor:
        assert x.shape[-1] == self.vectors.shape[-1]
        shape = x.shape[:-1]

        x = x.reshape(-1, 1, x.shape[-1])

        # Compute distances between z and vectors
        distances = torch.cdist(x, self.vectors.unsqueeze(0))
        # Get the closest vector
        _, indices = torch.min(distances, dim=-1)
        # Get the corresponding vector

        return indices.reshape(shape)

    @property
    def n_bins(self) -> int:
        return self.vectors.shape[0]

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        indices = self.codebook_lookup(x)
        return indices.long()


class LearnableVectorQuantization(VectorQuantization):
    def __init__(
        self,
        net: nn.Module,
        quantization_dim: int,
        n_bins: int,
    ):
        super().__init__(input_dim=quantization_dim, n_bins=n_bins)
        self.net = net

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.net(x)
        assert z.shape[-1] == self.vectors.shape[-1]
        return super().forward(z)

    def __repr__(self):
        s = f"{self.__class__.__name__}("
        s += f"\n  (net): " + self.net.__repr__().replace("\n", "\n  ")
        s += f"\n  (n_bins): {self.vectors.shape[0]}"
        s += "\n)"
        return s


class FixedQuantization(QuantizationFunction):
    def __init__(self, input_dim: int, thresholds: torch.Tensor):
        super().__init__()
        self.thresholds = thresholds
        self.input_dim = input_dim

    @property
    def n_bins(self) -> int:
        return (self.thresholds.shape[0] + 1) ** self.input_dim

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        bins = torch.bucketize(x, self.thresholds)
        flat_bins = (
            bins
            * (
                torch.FloatTensor([self.thresholds.shape[0] + 1])
                ** torch.arange(self.input_dim)
            ).to(bins.device)
        ).sum(-1)
        return flat_bins.long()



import torch
from torch import nn

CRITIC_TYPE = "critic_type"
SEPARABLE_CRITIC = "separable"
JOINT_CRITIC = "joint"
CRITIC_TYPES = [JOINT_CRITIC, SEPARABLE_CRITIC]


class Critic(nn.Module):
    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the value of the critic evaluated at the pair (x, y)
        :param x: a tensor representing x
        :param y: a tensor representing y
        :return: The value of the ratio estimator on the given pair
        """
        raise NotImplementedError()



from typing import Optional, Union

import torch
from torch.distributions import Distribution

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.estimators import MIEstimator
from torch_mist.utils.caching import (
    cached,
    reset_cache_after_call,
    reset_cache_before_call,
)


class GM(MIEstimator):
    def __init__(
        self,
        q_XY: JointDistribution,
        q_Y: Optional[Union[Distribution, JointDistribution]] = None,
        q_X: Optional[Union[Distribution, JointDistribution]] = None,
    ):
        super().__init__()
        self.q_XY = q_XY
        self._q_Y = q_Y
        self._q_X = q_X

    @property
    def q_X(self):
        if self._q_X:
            return self._q_X
        else:
            return self.q_XY.marginal("x")

    @property
    def q_Y(self):
        if self._q_Y:
            return self._q_Y
        else:
            return self.q_XY.marginal("y")

    @cached
    def approx_log_p_x(self, x: torch.Tensor) -> torch.Tensor:
        log_q_x = self.q_X.log_prob(x)
        assert (
            log_q_x.shape == x.shape[:-1]
            and not isinstance(x, torch.LongTensor)
        ) or (log_q_x.shape == x.shape and isinstance(x, torch.LongTensor))
        # The shape is [...]
        return log_q_x

    @cached
    def approx_log_p_y(self, y: torch.Tensor) -> torch.Tensor:
        log_q_y = self.q_Y.log_prob(y)
        assert (
            log_q_y.shape == y.shape[:-1]
            and not isinstance(y, torch.LongTensor)
        ) or (log_q_y.shape == y.shape and isinstance(y, torch.LongTensor))
        # The shape is [...]
        return log_q_y

    @cached
    def approx_log_p_xy(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        # The shape is [...]
        log_q_xy = self.q_XY.log_prob(x=x, y=y)

        assert (
            log_q_xy.shape == y.shape[:-1]
            and not isinstance(y, torch.LongTensor)
        ) or (log_q_xy.shape == y.shape and isinstance(y, torch.LongTensor))

        return log_q_xy

    @reset_cache_after_call
    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        log_q_xy = self.approx_log_p_xy(x=x, y=y)
        log_q_y = self.approx_log_p_y(y=y)
        log_q_x = self.approx_log_p_x(x=x)

        mi = log_q_xy - log_q_y - log_q_x

        assert (
            mi.shape == y.shape[:-1] and not isinstance(y, torch.LongTensor)
        ) or (mi.shape == y.shape and isinstance(y, torch.LongTensor))

        return mi

    @reset_cache_before_call
    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        log_q_xy = self.approx_log_p_xy(x=x, y=y)
        log_q_y = self.approx_log_p_y(y=y)
        log_q_x = self.approx_log_p_x(x=x)

        loss = -log_q_xy - log_q_y - log_q_x
        assert (
            loss.shape == y.shape[:-1] and not isinstance(y, torch.LongTensor)
        ) or (loss.shape == y.shape and isinstance(y, torch.LongTensor))

        return loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += "  (q_XY): " + str(self.q_XY).replace("\n", "\n  ") + "\n"
        if self.q_X is not None:
            s += "  (q_X): " + str(self.q_X).replace("\n", "\n  ") + "\n"
        if self.q_Y is not None:
            s += "  (q_Y): " + str(self.q_Y).replace("\n", "\n  ") + "\n"
        s += ")" + "\n"
        return s



import math
from abc import abstractmethod
from typing import Optional, List

import torch
from torch import nn


class Baseline(nn.Module):
    @abstractmethod
    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        raise NotImplementedError()


class ConstantBaseline(Baseline):
    def __init__(self, value: float = 0):
        super(ConstantBaseline, self).__init__()
        self.value = value

    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        return torch.zeros(f_.shape[1:]).to(f_.device) + self.value


class ExponentialMovingAverage(Baseline):
    def __init__(self, gamma: float = 0.9):
        super(ExponentialMovingAverage, self).__init__()
        self.ema = None
        assert 0 <= gamma <= 1
        self.gamma = gamma

    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        average = f_.exp().mean().detach()
        average = torch.clamp(average, min=1e-4, max=1e4)

        if self.ema is None:
            ema = average
            self.ema = ema
        else:
            if self.training:
                ema = self.gamma * self.ema + (1 - self.gamma) * average
                self.ema = ema
            else:
                ema = self.ema

        ma = ema.log()
        for _ in range(x.ndim - 1):
            ma = ma.unsqueeze(0)
        ma = ma.expand_as(x[..., 0])
        return ma


class BatchLogMeanExp(Baseline):
    def __init__(self, dims: str):
        assert dims in ["first", "all"]
        super(BatchLogMeanExp, self).__init__()
        self.dims = dims

    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # log 1/M \sum e^f_ = logsumexp(f_) - log M
        if self.dims == "all":
            f_ = f_.reshape(-1)

        M = f_.shape[0]
        b = torch.logsumexp(f_, 0) - math.log(M)

        if self.dims == "all":
            for _ in range(x.ndim - 1):
                b = b.unsqueeze(0)
            b = b.expand_as(x[..., 0])

        return b


class LearnableBaseline(Baseline):
    def __init__(self, net: nn.Module):
        super(LearnableBaseline, self).__init__()
        self.net = net

    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        return self.net(x).squeeze(-1)


class InterpolatedBaseline(Baseline):
    def __init__(
        self, baseline_1: Baseline, baseline_2: Baseline, alpha: float
    ):
        super(InterpolatedBaseline, self).__init__()
        assert 0 <= alpha <= 1
        self.alpha = alpha
        self.baseline_1 = baseline_1
        self.baseline_2 = baseline_2

    def forward(
        self,
        f_: torch.Tensor,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        b1 = self.baseline_1.forward(f_=f_, x=x, y=y)
        b2 = self.baseline_2.forward(f_=f_, x=x, y=y)

        assert b1.shape == b2.shape

        if self.alpha == 0:
            return b2
        elif self.alpha == 1:
            return b1
        else:
            # We use logsumexp for numerical stability
            # log alpha * exp(b1) + (1-alpha) * exp(b2) = log(exp(b1+log alpha) + exp(b2+log(1-alpha)))
            b = torch.logsumexp(
                torch.cat(
                    [
                        b1.unsqueeze(-1) + math.log(self.alpha),
                        b2.unsqueeze(-1) + math.log(1 - self.alpha),
                    ],
                    -1,
                ),
                -1,
            )

            return b


class AlphaTUBABaseline(InterpolatedBaseline):
    def __init__(self, x_dim: int, hidden_dims: List[int], alpha: float):
        from torch_mist.baseline.factories import baseline_nn

        baseline_1 = BatchLogMeanExp("first")
        baseline_2 = baseline_nn(x_dim, hidden_dims)
        super(AlphaTUBABaseline, self).__init__(
            baseline_1=baseline_1, baseline_2=baseline_2, alpha=alpha
        )



from typing import Type, Optional, Dict, Any, Union, Tuple, List

import torch
import pandas as pd
from torch.optim import Optimizer
from tqdm.autonotebook import tqdm
from torch.optim import Adam
from torch.optim.lr_scheduler import (
    SequentialLR,
    LinearLR,
    CosineAnnealingLR,
    ConstantLR,
    LRScheduler,
)
from torch.utils.data import random_split, DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device
from torch_mist.utils.data.dataset import SampleDataset
from torch_mist.utils.evaluation import evaluate_mi
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.logging.logger.pandas import PandasLogger
from torch_mist.utils.logging.logger.utils import instantiate_logger


def _instantiate_dataloaders(
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> Tuple[DataLoader, Optional[DataLoader]]:
    if (x is None) != (y is None):
        raise ValueError(
            "Either both x and y need to be specified or neither."
        )

    if not ((x is None) ^ (train_loader is None)):
        raise ValueError(
            "Either both x and y or the train_loader need to be specified."
        )
    if not (x is None):
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")

        # Make the dataloaders from the samples
        dataset = SampleDataset({"x": x, "y": y})
        if valid_percentage > 0:
            if not (valid_loader is None):
                raise ValueError(
                    "The valid_loader can't be specified when x and y are used"
                )

            n_valid = int(len(dataset) * valid_percentage)
            dataset, val_set = random_split(
                dataset, [len(dataset) - n_valid, n_valid]
            )
            valid_loader = DataLoader(
                val_set,
                batch_size=batch_size,
                num_workers=num_workers,
            )

        train_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
        )

    return train_loader, valid_loader


def _instantiate_optimizer(
    estimator: MIEstimator,
    max_epochs: int,
    iterations_per_epoch: int,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
) -> Tuple[Optimizer, LRScheduler]:
    params = [
        {"params": params}
        for params in estimator.parameters()
        if params.requires_grad
    ]

    if optimizer_params is None:
        optimizer_params = {"lr": 5e-4}

    opt = optimizer_class(params, **optimizer_params)

    # Cosine annealing with initial linear warmup
    if lr_annealing:
        max_iterations = iterations_per_epoch * max_epochs
        if not 0 <= warmup_percentage <= 1:
            raise ValueError("Warmup percentage must be between 0 and 1")
        warmup_iterations = int(
            iterations_per_epoch * max_epochs * warmup_percentage
        )
        lr_scheduler = SequentialLR(
            opt,
            [
                LinearLR(
                    opt, start_factor=1e-2, total_iters=warmup_iterations
                ),
                CosineAnnealingLR(
                    opt,
                    T_max=max_iterations - warmup_iterations,
                    eta_min=1e-5,
                ),
            ],
            milestones=[warmup_iterations],
        )
    else:
        lr_scheduler = ConstantLR(opt, 1.0)

    return opt, lr_scheduler


def _train_epoch(
    estimator: MIEstimator,
    train_loader: DataLoader,
    opt: Optimizer,
    lr_scheduler: LRScheduler,
    device: Union[str, torch.device],
    logger: Optional[Logger] = None,
    tqdm_iteration: Optional[tqdm] = None,
    fast_train: bool = False,
):
    for samples in train_loader:
        variables = unfold_samples(samples)
        variables = move_to_device(variables, device)

        loss = estimator.loss(**variables)

        # Compute the ratio only if necessary
        if not fast_train and not isinstance(logger, DummyLogger):
            estimator(**variables)

        opt.zero_grad()
        loss.backward()
        opt.step()
        lr_scheduler.step()
        logger.step()

        if tqdm_iteration:
            tqdm_iteration.update(1)
            tqdm_iteration.set_postfix_str(f"loss: {loss}")


def train_mi_estimator(
    estimator: MIEstimator,
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: int = 10,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    early_stopping: bool = True,
    patience: int = 3,
    delta: float = 0.001,
    fast_train: bool = False,
) -> Optional[Any]:
    train_loader, valid_loader = _instantiate_dataloaders(
        x=x,
        y=y,
        train_loader=train_loader,
        valid_loader=valid_loader,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    opt, lr_scheduler = _instantiate_optimizer(
        estimator=estimator,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        max_epochs=max_epochs,
        iterations_per_epoch=len(train_loader),
    )

    estimator.train()
    estimator = estimator.to(device)

    # If the logger is specified, we use it adding loss and mutual_information logs if not already specified
    # if it is None, use the default PandasLogger,
    # if False, instantiate a DummyLogger, which does not store any quantity
    logger = instantiate_logger(estimator, logger)

    best_mi = 0
    initial_patience = patience
    tqdm_epochs = (
        tqdm(total=max_epochs, desc="Epoch", position=1) if verbose else None
    )
    tqdm_iteration = (
        tqdm(total=len(train_loader), desc="Iteration", position=1)
        if verbose
        else None
    )

    for epoch in range(max_epochs):
        if tqdm_epochs:
            tqdm_iteration.reset()

        with logger.train():
            logger.new_epoch()
            _train_epoch(
                estimator=estimator,
                train_loader=train_loader,
                opt=opt,
                lr_scheduler=lr_scheduler,
                logger=logger,
                tqdm_iteration=tqdm_iteration,
                device=device,
                fast_train=fast_train,
            )

        if valid_loader is not None:
            with logger.valid():
                valid_mi = evaluate_mi(
                    estimator=estimator,
                    dataloader=valid_loader,
                    device=device,
                )

            if isinstance(valid_mi, dict):
                valid_mi = sum(valid_mi.values())

            if tqdm_epochs:
                tqdm_epochs.set_postfix_str(f"valid_mi: {valid_mi}")

            if (
                early_stopping
                and estimator.upper_bound
                or estimator.lower_bound
            ):
                improvement = (
                    (valid_mi - best_mi)
                    if estimator.upper_bound
                    else (best_mi - valid_mi)
                )
                if improvement >= delta:
                    # Improvement, update best and reset the patience
                    best_mi = valid_mi
                    patience = initial_patience
                else:
                    patience -= 1

                if patience < 0:
                    if verbose:
                        print("No improvements on validation, stopping.")
                    break

        if tqdm_epochs:
            tqdm_epochs.update(1)

    return logger.get_log()



from typing import Callable, TypeVar
import inspect

import torch

T = TypeVar("T")


def cached(method: Callable[..., T]) -> Callable[..., T]:
    def _is_cache_valid(self, key: str, **kwargs):
        assert key in self.__cache, f"Key {key} not in cache"
        if self.__cache[key] is None:
            return False
        stored_kwargs, stored_value = self.__cache[key]
        if stored_kwargs.keys() != kwargs.keys():
            return False

        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                if not torch.equal(stored_kwargs[k], v):
                    return False
            else:
                if stored_kwargs[k] != v:
                    return False

        return True

    def cached_method(self, *args, **kwargs):
        # transform args in kwargs
        keys = inspect.signature(method).parameters.keys()
        unused_keys = [key for key in keys if key not in kwargs]
        if "self" in unused_keys:
            unused_keys.remove("self")
        new_kwargs = {unused_keys[i]: arg for i, arg in enumerate(args)}
        kwargs = {**kwargs, **new_kwargs}

        if not hasattr(self, "__cache"):
            self.__cache = {}
            self.__caching_enabled = True

        if not self.__caching_enabled:
            return method(self, *args, **kwargs)

        key = method.__name__
        if key not in self.__cache or not _is_cache_valid(self, key, **kwargs):
            value = method(self, **kwargs)
            # print(f"Cache miss for {key}")
            self.__cache[key] = (kwargs, value)
        else:
            # print(f"Cache hit for {key}")
            value = self.__cache[key][1]
        return value

    return cached_method


def reset_cache_after_call(method: Callable[..., T]) -> Callable[..., T]:
    def reset_cache(self, *args, **kwargs):
        value = method(self, *args, **kwargs)
        if hasattr(self, "__cache"):
            self.__cache = {key: None for key in self.__cache.keys()}
        return value

    return reset_cache


def reset_cache_before_call(method: Callable[..., T]) -> Callable[..., T]:
    def reset_cache(self, *args, **kwargs):
        if hasattr(self, "__cache"):
            self.__cache = {key: None for key in self.__cache.keys()}
        return method(self, *args, **kwargs)

    return reset_cache



from abc import abstractmethod
from typing import Any, Callable, Dict, Optional, ContextManager, Union, List

SPLITS = ["train", "valid", "test"]


class LoggingContext:
    def __init__(self, logger: Any, context: str):
        self.logger = logger
        self.context = context

    def __enter__(self) -> "LoggingContext":
        assert self.logger._context is None
        self.logger._context = self.context

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger._context = None


class Logger:
    def __init__(self):
        self.iteration = 0
        self.epoch = 0
        self._global_info = {}
        self._logged_methods = set()
        self._context = None

    def train(self) -> ContextManager[LoggingContext]:
        return LoggingContext(self, "train")

    def valid(self) -> ContextManager[LoggingContext]:
        return LoggingContext(self, "valid")

    def test(self) -> ContextManager[LoggingContext]:
        return LoggingContext(self, "test")

    def set_global_info(self, **kwargs):
        self._global_info = kwargs

    def log_method(
        self,
        instance: Any,
        method_name: str,
        agg: Optional[
            Callable[[Any], Union[List[Dict[str, Any]], Dict[str, Any]]]
        ] = None,
    ):
        assert method_name not in self._logged_methods
        self._logged_methods.add(method_name)

        original_method = getattr(instance, method_name)

        if agg is None:

            def mean(x):
                return {"value": x.mean().item(), "metric": "mean"}

            agg = mean

        def method_with_hook(*args, **kwargs):
            value = original_method(*args, **kwargs)
            if self._context:
                stats = agg(value.detach())
                if not (stats is None):
                    if isinstance(stats, list):
                        for stat in stats:
                            assert "metric" in stat
                            assert "value" in stat
                            stat["quantity"] = method_name
                            self.log(stat)
                    else:
                        assert isinstance(stats, dict)
                        assert "metric" in stats
                        assert "value" in stats
                        stats["quantity"] = method_name
                        self.log(stats)

            return value

        setattr(instance, method_name, method_with_hook)

    def is_logged(self, method_name: str) -> bool:
        return method_name in self._logged_methods

    @abstractmethod
    def _log(self, data: Dict[str, Any]):
        raise NotImplementedError()

    def log(self, data: Dict[str, Any]):
        data["iteration"] = self.iteration
        data["epoch"] = self.epoch
        data["split"] = self._context
        data.update(self._global_info)
        self._log(data)

    def step(self):
        assert (
            self._context == "train"
        ), "Iterations can be updated only in train context"
        self.iteration += 1

    def new_epoch(self):
        assert (
            self._context == "train"
        ), "Iterations can be updated only in train context"
        self.epoch += 1

    @abstractmethod
    def _reset(self):
        raise NotImplementedError()

    def reset(self):
        self.iteration = 0
        self.epoch = 0
        self._global_info = {}
        self._reset()

    def get_log(self) -> Optional[Any]:
        return None


class DummyLogger(Logger):
    def _log(self, data: Dict[str, Any]):
        pass

    def _reset(self):
        pass



from abc import abstractmethod

import torch
from pyro.distributions import ConditionalDistribution
from torch.distributions import Distribution

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.caching import (
    cached,
)


class GenerativeMIEstimator(MIEstimator):
    @cached
    def approx_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        raise NotImplementedError()

    @abstractmethod
    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def log_ratio(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        xdim = x.ndim + (1 if isinstance(x, torch.LongTensor) else 0)
        ydim = y.ndim + (1 if isinstance(y, torch.LongTensor) else 0)
        assert xdim == ydim, f"x.ndim={xdim}, y.ndim={ydim}"

        # Compute the ratio using the primal KL bound
        approx_log_p_y_given_x = self.approx_log_p_y_given_x(x=x, y=y)
        approx_log_p_y = self.approx_log_p_y(x=x, y=y)

        assert (
            approx_log_p_y_given_x.ndim == approx_log_p_y.ndim == x.ndim - 1
        ), f"log_p_y_x.ndim={approx_log_p_y_given_x.ndim}, log_p_y.ndim={approx_log_p_y.ndim}"
        log_ratio = approx_log_p_y_given_x - approx_log_p_y

        return log_ratio


class ConditionalGenerativeMIEstimator(GenerativeMIEstimator):
    def __init__(self, q_Y_given_X: ConditionalDistribution):
        super().__init__()
        self.q_Y_given_X = q_Y_given_X

    @cached
    def q_Y_given_x(self, x: torch.Tensor) -> Distribution:
        # q(Y|X=x)
        q_Y_given_x = self.q_Y_given_X.condition(x)

        return q_Y_given_x

    @cached
    def approx_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        q_Y_given_x = self.q_Y_given_x(x=x)

        # Compute log q(Y=y|X=x)]
        log_q_y_given_x = q_Y_given_x.log_prob(y)

        assert (
            log_q_y_given_x.shape == y.shape[:-1]
            and not isinstance(y, torch.LongTensor)
        ) or (
            log_q_y_given_x.shape == y.shape
            and isinstance(y, torch.LongTensor)
        ), f"log_p_Y_X.shape={log_q_y_given_x.shape}, y.shape={y.shape}"

        return log_q_y_given_x

    @abstractmethod
    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  "
            + "(q_Y_given_X): "
            + str(self.q_Y_given_X).replace("\n", "  \n")
            + "\n"
        )
        s += ")" + "\n"

        return s



from typing import List, Tuple, Dict

import torch
from torch import nn

from torch_mist.estimators.base import MIEstimator


class MultiMIEstimator(MIEstimator):
    def __init__(self, estimators: Dict[Tuple[str, str], MIEstimator]):
        super().__init__()

        self.estimators = nn.ModuleDict()
        for (x_key, y_key), estimator in estimators.items():
            key = f"{x_key};{y_key}"
            self.estimators[key] = estimator

    def broadcast_function(
        self, function_name: str, **variables
    ) -> Dict[Tuple[str, str], torch.Tensor]:
        values = {}
        for key, estimator in self.estimators.items():
            x_key, y_key = key.split(";")
            if not (x_key in variables):
                raise ValueError(f"The variable {x_key} is not specified.")
            if not (y_key in variables):
                raise ValueError(f"The variable {y_key} is not specified.")

            x = variables[x_key]
            y = variables[y_key]

            value = getattr(estimator, function_name)(x=x, y=y)
            values[(x_key, y_key)] = value
        return values

    def loss(self, **variables) -> torch.Tensor:
        losses = self.broadcast_function("loss", **variables)
        total_loss: torch.Tensor = 0
        for loss in losses.values():
            total_loss += loss
        return total_loss

    def mutual_information(
        self, **variables
    ) -> Dict[Tuple[str, str], torch.Tensor]:
        return self.broadcast_function("expected_log_ratio", **variables)

    def log_ratio(self, **variables) -> Dict[Tuple[str, str], torch.Tensor]:
        return self.broadcast_function("log_ratio", **variables)

    def forward(self, **variables) -> Dict[Tuple[str, str], torch.Tensor]:
        return self.mutual_information(**variables)



from typing import Optional, Union

from torch import nn

from torch_mist.utils.logging.logger.base import DummyLogger, Logger
from torch_mist.utils.logging.logger.pandas import PandasLogger


def instantiate_logger(
    model: nn.Module, logger: Optional[Union[bool, Logger]]
):
    # if logger is unspecified, use the default PandasLogger
    if logger is None:
        logger = PandasLogger()

    # if the logger is not logging anything, log the value of mutual information and loss by default
    if logger and not isinstance(logger, DummyLogger):
        # If no method is logged, add the loss and the mutual_information when available
        if len(logger._logged_methods) == 0:
            if hasattr(model, "mutual_information"):
                logger.log_method(model, "mutual_information")
            if hasattr(model, "loss"):
                logger.log_method(model, "loss")
    # if logger is False, we instantiate a dummy logger which does not log.
    else:
        logger = DummyLogger()

    return logger



from typing import Optional

import torch

from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.discriminative.base import DiscriminativeMIEstimator
from torch_mist.estimators.generative.base import GenerativeMIEstimator
from torch_mist.utils.freeze import is_frozen


class HybridMIEstimator(MIEstimator):
    def __init__(
        self,
        generative_estimator: Optional[GenerativeMIEstimator] = None,
        discriminative_estimator: Optional[DiscriminativeMIEstimator] = None,
    ):
        super().__init__()
        self.generative_estimator = generative_estimator
        self.discriminative_estimator = discriminative_estimator

    def unnormalized_log_ratio(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        log_ratio = self.generative_estimator.log_ratio(x, y)
        log_ratio += self.discriminative_estimator.unnormalized_log_ratio(x, y)
        return log_ratio

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        log_ratio = self.generative_estimator.log_ratio(x, y)
        proposal = self.generative_estimator.q_Y_given_x(x=x)
        self.discriminative_estimator.proposal = proposal
        log_ratio += self.discriminative_estimator.log_ratio(x, y)
        return log_ratio

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        if is_frozen(self.generative_estimator):
            e1 = 0.0
        else:
            e1 = self.generative_estimator.batch_loss(x, y)
        if is_frozen(self.discriminative_estimator):
            e2 = 0.0
        else:
            proposal = self.generative_estimator.q_Y_given_x(x=x)
            self.discriminative_estimator.proposal = proposal
            e2 = self.discriminative_estimator.batch_loss(x, y)
        return e1 + e2



import torch

from torch import nn
from torch.distributions import Distribution, Categorical

from torch_mist.distributions.parametrizations.map import LogitsMap
from torch_mist.distributions.transforms import ConditionalDistributionModule


class CategoricalModule(Distribution, nn.Module):
    def __init__(
        self,
        logits: torch.tensor,
        temperature: float = 1.0,
        learnable: bool = False,
    ):
        nn.Module.__init__(self)
        Distribution.__init__(
            self,
            event_shape=torch.Size([logits.shape[0]]),
            validate_args=False,
        )
        if learnable:
            self.logits = nn.Parameter(logits)
        else:
            self.register_buffer("logits", logits)
        self.parametrization = LogitsMap()
        self.temperature = temperature

    def rsample(self, sample_shape=torch.Size()):
        params = self.parametrization([self.logits / self.temperature])
        return Categorical(**params).rsample(sample_shape)

    def log_prob(self, value):
        params = self.parametrization([self.logits / self.temperature])
        return Categorical(**params).log_prob(value)

    def __repr__(self):
        return "Categorical()"


class ConditionalCategoricalModule(ConditionalDistributionModule):
    def __init__(self, net: nn.Module, temperature: float = 1.0):
        super(ConditionalCategoricalModule, self).__init__()
        self.net = net
        self.temperature = temperature
        self.parametrization = LogitsMap()

    def condition(self, x):
        return Categorical(
            **self.parametrization(self.net(x) / self.temperature)
        )



from typing import Optional, List, Dict

import torch
from torch.distributions import Distribution

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.utils.shape import expand_to_same_shape


class TorchJointDistribution(JointDistribution):
    def __init__(
        self,
        torch_dist: Distribution,
        variables: List[str],
        splits: Optional[List[int]] = None,
        split_dim: int = -1,
        name: Optional[str] = "p",
    ):
        super().__init__(variables=variables, name=name)
        self.torch_dist = torch_dist
        if not (splits is None):
            assert len(splits) == len(variables)
        self.splits = splits
        self.split_dim = split_dim

    def _tensor_to_dict(self, tensor: torch.Tensor) -> Dict[str, torch.Tensor]:
        if self.splits:
            tensors = torch.split(tensor, self.splits, dim=self.split_dim)
        else:
            tensors = torch.split(tensor, [1, 1], dim=self.split_dim)
            tensors = [tensor.squeeze(-1) for tensor in tensors]
        return {name: value for name, value in zip(self.variables, tensors)}

    def _dict_to_tensor(self, dict: Dict[str, torch.Tensor]) -> torch.Tensor:
        if self.splits is None:
            dict = {
                variable: tensor.unsqueeze(self.split_dim)
                for variable, tensor in dict.items()
            }
        return torch.cat(
            [dict[variable] for variable in self.variables], self.split_dim
        )

    def _log_prob(self, **kwargs) -> torch.Tensor:
        if len(self.variables) == 1:
            variable = self.variables[0]
            assert len(kwargs) == 1 and variable in kwargs
            tensor = kwargs[variable]
        else:
            assert len(kwargs) == len(
                self.variables
            ), f"passed: {kwargs.keys()} != required: {self.variables}"

            for label in kwargs:
                assert label in self.variables

            variables = []
            tensors = []
            for variable, tensor in kwargs.items():
                tensors.append(tensor)
                variables.append(variable)

            tensors = expand_to_same_shape(*tensors)
            tensor = self._dict_to_tensor(
                {variables[i]: tensors[i] for i in range(len(variables))}
            )

        # Compute the log prob
        log_prob = self.torch_dist.log_prob(tensor)
        return log_prob

    def _entropy(self, variables: List[str]) -> torch.Tensor:
        if len(variables) == 1 and len(self.variables) == 1:
            return self.torch_dist.entropy()
        else:
            raise NotImplementedError()

    def sample(
        self, sample_shape: torch.Size = torch.Size()
    ) -> Dict[str, torch.Tensor]:
        sample = self.torch_dist.sample(sample_shape)
        if len(self.variables) == 1:
            variable = self.variables[0]
            return {variable: sample}
        else:
            return self._tensor_to_dict(sample)

    def rsample(
        self, sample_shape: torch.Size = torch.Size()
    ) -> Dict[str, torch.Tensor]:
        rsample = self.torch_dist.rsample(sample_shape)
        if len(self.variables) == 1:
            variable = self.variables[0]
            return {variable: rsample}
        else:
            return self._tensor_to_dict(rsample)



from typing import Optional, Any, Union, Type, Dict, Tuple

import torch
import numpy as np
import pandas as pd
from torch.optim import Optimizer, Adam

from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.factories import instantiate_estimator
from torch_mist.utils.logging.logger.base import Logger
from torch_mist.utils.logging.logger.utils import instantiate_logger
from torch_mist.utils.train.mi_estimator import train_mi_estimator
from torch_mist.utils.evaluation import evaluate_mi


def estimate_mi(
    estimator_name: str,
    x: Optional[Union[torch.Tensor, np.array]] = None,
    y: Optional[Union[torch.Tensor, np.array]] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    test_loader: Optional[Any] = None,
    valid_percentage: float = 0.2,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: int = 10,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    batch_size: Optional[int] = 64,
    evaluation_batch_size: Optional[int] = None,
    num_workers: int = 8,
    early_stopping: bool = True,
    patience: int = 3,
    delta: float = 0.001,
    return_estimator: bool = False,
    fast_train: bool = False,
    **kwargs,
) -> Union[
    float,
    Tuple[float, pd.DataFrame],
    Tuple[float, MIEstimator],
    Tuple[float, MIEstimator, pd.DataFrame],
]:
    if not (x is None):
        if len(x.shape) == 1:
            x = x.reshape(-1, 1)
    if not (y is None):
        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

    if "x_dim" not in kwargs and not (x is None):
        kwargs["x_dim"] = x.shape[-1]
    if "y_dim" not in kwargs and not (y is None):
        kwargs["y_dim"] = y.shape[-1]

    if verbose:
        print(f"Instantiating the {estimator_name} estimator")
    estimator = instantiate_estimator(estimator_name=estimator_name, **kwargs)
    if verbose:
        print(estimator)
        print("Training the estimator")

    if evaluation_batch_size is None:
        evaluation_batch_size = batch_size

    # If the logger is specified, we use it, if it is None, use the PandasLogger, if false, instantiate a DummyLogger
    logger = instantiate_logger(estimator, logger)

    train_log = train_mi_estimator(
        estimator=estimator,
        x=x,
        y=y,
        train_loader=train_loader,
        valid_loader=valid_loader,
        valid_percentage=valid_percentage,
        device=device,
        max_epochs=max_epochs,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        verbose=verbose,
        logger=logger,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        batch_size=batch_size,
        early_stopping=early_stopping,
        patience=patience,
        delta=delta,
        num_workers=num_workers,
        fast_train=fast_train,
    )

    if verbose:
        print("Evaluating the value of Mutual Information")
    if not (test_loader is None):
        x = None
        y = None
    elif x is None:
        print(
            "Warning: using the train_loader to estimate the value of mutual information. Please specify a test_loader"
        )
        test_loader = train_loader

    with logger.test():
        mi_value = evaluate_mi(
            estimator=estimator,
            x=x,
            y=y,
            dataloader=test_loader,
            batch_size=evaluation_batch_size,
            device=device,
            num_workers=num_workers,
        )

    if not (train_log is None) and return_estimator:
        return mi_value, estimator, train_log
    elif not (train_log is None):
        return mi_value, train_log
    elif return_estimator:
        return mi_value, estimator
    else:
        return mi_value



import torch
from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.implementations.l1out import L1Out


class CLUB(L1Out):
    def __init__(
        self, q_Y_given_X: ConditionalDistribution, neg_samples: int = 0
    ):
        super().__init__(q_Y_given_X=q_Y_given_X, neg_samples=neg_samples)

    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # log prob of the samples in the batch, shape [M, N, ...]
        log_p_y_given_x = self._broadcast_log_p_y_given_x(x, y)

        # Compute the expectation over M. The shape is [N, ...]
        log_p_y = torch.mean(log_p_y_given_x, dim=0)

        assert log_p_y.shape == y.shape[:-1]

        return log_p_y



import torch

from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    DiscriminativeMIEstimator,
)
from torch_mist.utils.caching import cached


class FLO(DiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        amortized_critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
        )
        self.amortized_critic = amortized_critic

    @cached
    def amortized_estimate(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        return self.amortized_critic(x=x, y=y)

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        assert x.ndim == y.ndim
        # Approximate the log-ratio p(y|x)/p(y) on samples from p(x,y).
        # x and y have shape [..., X_DIM] and [..., Y_DIM] respectively

        # Evaluate the unnormalized_log_ratio f(x,y) on the samples from p(x, y), with shape [...]
        f = self.unnormalized_log_ratio(x, y)

        # Evaluate the unnormalized_log_ratio f(x,y) on the samples from r(x, y), with shape [M, ...]
        f_ = self.critic_on_negatives(x, y)

        assert f_.shape[1:] == f.shape

        # Compute the negative amortized log ratio. It has shape [...]
        u = self.amortized_estimate(x=x, y=y)
        assert (
            u.ndim == f_.ndim - 1
        ), f"Baseline has ndim {u.ndim} while f_ has ndim {f_.ndim}"

        loss = -u + (torch.logsumexp(f_, 0) - f + u).exp() / f_.shape[0] + 1

        return loss

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.amortized_estimate(x=x, y=y)

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (critic): " + str(self.critic).replace("\n", "\n" + "  ") + "\n"
        )
        s += (
            "  (amortized_critic): "
            + str(self.amortized_critic).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s



import collections
from typing import Optional, Any, List, Union, Dict

import torch
import numpy as np
from torch.utils.data import DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device
from torch_mist.utils.data.dataset import SampleDataset


def evaluate_mi(
    estimator: MIEstimator,
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    dataloader: Optional[Any] = None,
    device: torch.device = torch.device("cpu"),
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> Union[float, Dict[str, float]]:
    if (x is None) != (y is None):
        raise ValueError(
            "Either both x and y need to be specified or neither."
        )
    if (
        (x is None and y is None and dataloader is None)
        or not (x is None)
        and not (y is None)
        and not (dataloader is None)
    ):
        raise ValueError(
            "Either both x and y or the train_loader need to be specified."
        )
    if not (x is None) and not (y is None):
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")

        # Make a train_loader from the samples
        dataset = SampleDataset({"x": x, "y": y})
        dataloader = DataLoader(
            dataset, batch_size=batch_size, num_workers=num_workers
        )

    estimator.eval()
    estimator = estimator.to(device)

    mis = collections.defaultdict(list)
    unfold = False
    for samples in dataloader:
        variables = unfold_samples(samples)
        variables = move_to_device(variables, device)

        estimation = estimator(**variables)

        if isinstance(estimation, dict):
            for (x_key, y_key), value in estimation.items():
                mis[f"I({x_key};{y_key})"].append(value.item())
        else:
            mis["I(x;y)"].append(estimation.item())
            unfold = True

    mi = {}
    for name, values in mis.items():
        mi[name] = np.mean(values)

    if unfold:
        return mi["I(x;y)"]
    return mi



from typing import Dict, Sequence
import torch

from torch.utils.data import Dataset


class SampleDataset(Sequence, Dataset):
    def __init__(self, samples: Dict[str, torch.Tensor]):
        self.samples = samples
        # Check all the samples have the same length
        lengths = [len(value) for value in samples.values()]
        assert all([length == lengths[0] for length in lengths])
        self.n_samples = lengths[0]

    def __getitem__(self, item):
        return {
            name: value[item]
            if value[item].ndim == 1
            else value[item].reshape(1)
            for name, value in self.samples.items()
        }

    def __len__(self):
        return self.n_samples



from typing import Optional, Iterator, Dict, Any, Union

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader

from torch_mist.utils.data.dataset import SampleDataset
from torch_mist.utils.logging.logger.base import Logger
from torch_mist.utils.logging.logger.utils import instantiate_logger


def train_model(
    model: nn.Module,
    max_epochs: int,
    dataloader: Optional[Iterator] = None,
    data: Optional[Union[torch.Tensor, np.array]] = None,
    optimizer_class=torch.optim.Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
    logger: Optional[Union[bool, Logger]] = None,
) -> Optional[Any]:
    if (data is None and dataloader is None) or (
        not (data is None) and not (dataloader is None)
    ):
        raise ValueError(
            "Either a set of samples or a dataloader need to be specified."
        )
    if not (data is None):
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")
        dataset = SampleDataset({"x": data})
        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
        )

    if optimizer_params is None:
        optimizer_params = dict(lr=1e-3)

    opt = optimizer_class(model.parameters(), **optimizer_params)
    # If the logger is specified, we use it adding loss and mutual_information logs if not already specified
    # if it is None, use the default PandasLogger,
    # if False, instantiate a DummyLogger, which does not store any quantity
    logger = instantiate_logger(model, logger)

    with logger.train():
        logger.new_epoch()
        for epoch in range(max_epochs):
            for samples in dataloader:
                if isinstance(samples, dict):
                    if "x" not in samples:
                        raise ValueError("Expected 'x' key in samples")
                    data = samples["x"]
                elif isinstance(samples, torch.Tensor):
                    data = samples
                else:
                    raise ValueError(
                        f"Unknown sample type {type(samples)}, expected dict containing keys 'x' and 'y' or torch.Tensor"
                    )

                opt.zero_grad()
                model.loss(data).backward()
                opt.step()
                logger.step()

    train_log = logger.get_log()
    return train_log



from typing import Dict, List

import torch
from torch.distributions import (
    MultivariateNormal,
    Normal,
    Independent,
    Distribution,
)

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.distributions.joint.wrapper import TorchJointDistribution


class JointMultivariateNormal(TorchJointDistribution):
    def __init__(
        self,
        n_dim: int,
        rho: float = 0.9,
        sigma: float = 1,
        device: torch.device = torch.device("cpu"),
    ):
        loc = torch.zeros(n_dim, 2).to(device)
        covariance = torch.eye(2).to(device)
        covariance[0, 1] = covariance[1, 0] = rho
        covariance = covariance * sigma
        covariance = covariance.unsqueeze(0).repeat(n_dim, 1, 1)

        p_XY = Independent(MultivariateNormal(loc, covariance), 1)

        self.p_X = Independent(Normal(torch.zeros(n_dim).to(device), sigma), 1)

        super().__init__(torch_dist=p_XY, variables=["x", "y"])

    def _marginal(self, variables: List[str]) -> JointDistribution:
        assert len(variables) == 1
        variable = variables[0]

        return TorchJointDistribution(
            torch_dist=self.p_X,
            variables=[variable],
        )

    def _entropy(self, variables: List[str]) -> torch.Tensor:
        if len(variables) == 2:
            return self.torch_dist.entropy()
        else:
            return self.p_X.entropy()



from typing import Optional

import torch

from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.caching import (
    reset_cache_after_call,
    reset_cache_before_call,
)


class BA(ConditionalGenerativeMIEstimator):
    upper_bound: bool = True

    def __init__(
        self,
        q_Y_given_X: ConditionalDistribution,
        entropy_y: Optional[torch.Tensor] = None,
    ):
        super().__init__(q_Y_given_X=q_Y_given_X)
        if not isinstance(entropy_y, torch.Tensor):
            entropy_y = torch.tensor(entropy_y)
        entropy_y = entropy_y.squeeze()
        assert entropy_y.ndim == 0
        self.register_buffer("entropy_y", entropy_y)

    @reset_cache_after_call
    def mutual_information(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        entropy_y_given_x = -self.approx_log_p_y_given_x(x=x, y=y).mean()
        return self.entropy_y - entropy_y_given_x

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -self.approx_log_p_y_given_x(x=x, y=y)



from typing import Optional

import torch

from torch_mist.baseline import BatchLogMeanExp
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.implementations.js import JS
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
    CombinedDiscriminativeMIEstimator,
)


class SMILE(CombinedDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
        tau: float = 5.0,
    ):
        super().__init__(
            train_estimator=JS(critic=critic, neg_samples=neg_samples),
            eval_estimator=BaselineDiscriminativeMIEstimator(
                critic=critic,
                neg_samples=neg_samples,
                baseline=BatchLogMeanExp("all"),
            ),
        )
        self.tau = tau

    def batch_approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
    ) -> torch.Tensor:
        return super().batch_approx_log_partition(
            x=x, y=y, f_=torch.clamp(f_, min=-self.tau, max=self.tau)
        )

    def loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> Optional[torch.Tensor]:
        return JS.loss(self, x, y)



from abc import ABC
from typing import List, Union, Dict

import torch
from torch import nn

from pyro.distributions import (
    ConditionalDistribution,
    ConditionalTransform,
    ConditionalTransformedDistribution,
)
from torch.distributions import (
    Distribution,
    Transform,
    TransformedDistribution,
)


class DistributionModule(Distribution, nn.Module, ABC):
    def __init__(self, validate_args: bool = False):
        Distribution.__init__(self, validate_args=validate_args)
        nn.Module.__init__(self)

    def __repr__(self):
        return nn.Module.__repr__(self)


class ConditionalDistributionModule(ConditionalDistribution, nn.Module, ABC):
    def __init__(self):
        ConditionalDistribution.__init__(self)
        nn.Module.__init__(self)


class ConditionalTransformedDistributionModule(
    ConditionalTransformedDistribution, nn.Module
):
    def __init__(
        self,
        base_dist: Union[ConditionalDistribution, Distribution],
        transforms: Union[
            ConditionalTransform,
            List[ConditionalTransform],
            Dict[str, Union[ConditionalTransform, Transform]],
            Transform,
            List[Transform],
            None,
        ],
    ):
        if isinstance(transforms, ConditionalTransform) or isinstance(
            transforms, Transform
        ):
            transforms = [transforms]
        elif isinstance(transforms, dict):
            transforms = [transforms[k] for k in sorted(transforms.keys())]

        if transforms is not None:
            transforms = nn.ModuleList(list(transforms))
        else:
            transforms = []

        self._base_dist_repr = base_dist.__repr__()

        nn.Module.__init__(self)
        self.base_dist = base_dist
        self.transforms = transforms

    def condition(self, context):
        base_dist = (
            self.base_dist.condition(context)
            if isinstance(self.base_dist, ConditionalDistribution)
            else self.base_dist
        )
        transforms = [
            (
                t.condition(context)
                if isinstance(t, ConditionalTransform)
                else t
            )
            for t in self.transforms
        ]

        return TransformedDistribution(base_dist, transforms)

    def clear_cache(self):
        pass

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "\n  (base_dist): " + str(self._base_dist_repr).replace(
            "\n", "\n  "
        )
        s += "\n  (transforms): " + str(self.transforms).replace("\n", "\n  ")
        s += "\n"
        return s


class TransformedDistributionModule(DistributionModule):
    def __init__(
        self,
        base_dist: Distribution,
        transforms: Union[
            Transform, List[Transform], Dict[str, Transform], None
        ],
    ):
        super().__init__()

        self.base_dist = base_dist

        if isinstance(transforms, ConditionalTransform) or isinstance(
            transforms, Transform
        ):
            transforms = [transforms]
        elif hasattr(transforms, "keys"):
            transforms = [transforms[k] for k in sorted(transforms.keys())]

        self.transforms = nn.ModuleList(transforms)

    def rsample(self, sample_shape=torch.Size()):
        return TransformedDistribution(
            base_distribution=self.base_dist,
            transforms=[t for t in self.transforms],
        ).rsample(sample_shape)

    def log_prob(self, value):
        return TransformedDistribution(
            base_distribution=self.base_dist,
            transforms=[t for t in self.transforms],
            validate_args=False,
        ).log_prob(value)

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "\n  (base_dist): " + str(self.base_dist).replace("\n", "\n  ")
        s += "\n  (transforms): " + str(self.transforms).replace("\n", "\n  ")
        s += "\n)"
        return s



from pyro.distributions import ConditionalDistribution

from torch_mist.distributions.joint.categorical import JointCategorical
from torch_mist.estimators import DoE
from torch_mist.estimators.transformed.base import TransformedMIEstimator
from torch_mist.quantization.functions import QuantizationFunction
from torch_mist.utils.freeze import freeze


class PQ(TransformedMIEstimator):
    def __init__(
        self,
        q_QY_given_X: ConditionalDistribution,
        Q_y: QuantizationFunction,
        temperature: float = 1.0,
    ):
        freeze(Q_y)

        super().__init__(
            transforms={"y": Q_y},
            base_estimator=DoE(
                q_Y_given_X=q_QY_given_X,
                q_Y=JointCategorical(
                    variables=["y"],
                    bins=[Q_y.n_bins],
                    temperature=temperature,
                ),
            ),
        )



from typing import Optional

from torch_mist.distributions.joint.categorical import JointCategorical
from torch_mist.estimators import GM
from torch_mist.estimators.transformed.base import TransformedMIEstimator
from torch_mist.quantization.functions import QuantizationFunction
from torch_mist.utils.freeze import freeze


class BinnedMIEstimator(TransformedMIEstimator):
    # Technically this is not a lower-bound but the estimation of marginal entropy is usually accurate
    lower_bound = True

    def __init__(
        self,
        Q_x: Optional[QuantizationFunction] = None,
        Q_y: Optional[QuantizationFunction] = None,
        temperature: float = 1.0,
    ):
        freeze(Q_x)
        freeze(Q_y)

        q_XY = JointCategorical(
            variables=["x", "y"],
            bins=[Q_x.n_bins, Q_y.n_bins],
            temperature=temperature,
            name="q",
        )

        super().__init__(
            base_estimator=GM(
                q_XY=q_XY,
            ),
            transforms={
                "x": Q_x,
                "y": Q_y,
            },
        )



from torch.utils.data import Sampler
from typing import Union, Optional
import torch
import numpy as np


class SameAttributeSampler(Sampler):
    def __init__(
            self,
            batch_size: int,
            attributes: Optional[Union[torch.Tensor, np.ndarray]] = None,
            n_samples: Optional[int] = None,
            min_batch_size: int = 0,
    ):
        super(SameAttributeSampler, self).__init__(attributes)

        if attributes is not None and n_samples is not None:
            assert len(attributes) == n_samples, "Attributes must be of length n_samples"
        assert attributes is not None or n_samples is not None, "Either attributes or n_samples must be provided"
        if attributes is not None:
            if isinstance(attributes, torch.Tensor):
                attributes = attributes.data.numpy()
            if attributes.ndim == 1:
                attributes = attributes.reshape(-1, 1)
        if n_samples is None:
            n_samples = len(attributes)

        self.n_samples = n_samples

        self._attributes = attributes
        self.batch_size = batch_size
        self.all_ids = np.arange(n_samples)
        self._len = None
        self._cache = dict()
        self.min_batch_size = min_batch_size

    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        self._attributes = attributes
        self._cache = dict()
        self._len = None

    def _compute_len(self):

        mask = np.ones(len(self.attributes)) == 1
        l = 0
        while mask.sum() > 0:
            batch_ids = self.make_batch(mask)
            # Make sure the selected ids are not re-sampled for this epoch
            mask[batch_ids] = False
            if len(batch_ids) >= self.min_batch_size:
                l += 1
        self._len = l

    def compare_attributes(self, a):
        if isinstance(a, torch.Tensor):
            a = a.data.numpy()
        a_key = a.tobytes()
        if not (a_key in self._cache):
            v = np.sum(self.attributes == a, -1) == self.attributes.shape[-1]
            self._cache[a_key] = v
        return self._cache[a_key]

    def make_batch(self, mask: np.ndarray) -> np.ndarray:
        if self.attributes is not None:
            # Pick a random id that has not been sampled yet
            a_idx = np.random.choice(self.all_ids[mask], 1)[0]

            a = self.attributes[a_idx]

            # Consider all the ids with the same attribute that have not been sampled
            a_mask = self.compare_attributes(a)
            mask = mask & a_mask

        viable_ids = self.all_ids[mask]

        # If the number of available ids is at least as big as the batch size, select batch_size of them at random
        if self.batch_size <= len(viable_ids):
            batch_ids = np.random.choice(viable_ids, self.batch_size, replace=False)
        else:
            # Otherwise, select all the ids that are available without repetition (smaller batch)
            batch_ids = viable_ids

        return batch_ids

    def __iter__(self):
        mask = np.ones(self.n_samples) == 1

        while mask.sum() > 0:
            batch_ids = self.make_batch(mask)
            # Make sure the selected ids are not re-sampled for this epoch
            mask[batch_ids] = False
            # Discard batches that are too small
            if len(batch_ids) >= self.min_batch_size:
                yield batch_ids

    def __len__(self):
        if self.attributes is None:
            extra_batch = 1 if self.n_samples % self.batch_size >= self.min_batch_size else 0
            return self.n_samples // self.batch_size + extra_batch

        if self._len is None:
            self._compute_len()
        return self._len


from typing import List, Optional

from pyro.nn import DenseNN
from torch import nn

from torch_mist.critic.separable import SeparableCritic
from torch_mist.critic.joint import JointCritic
from torch_mist.critic.base import CRITIC_TYPES, JOINT_CRITIC
from torch_mist.nn import Normalize

SYMMETRIC_HEADS = "symmetric"
ASYMMETRIC_HEADS = "asymmetric"
ONE_HEAD = "one"
POSSIBLE_HEADS = [SYMMETRIC_HEADS, ASYMMETRIC_HEADS, ONE_HEAD]


def separable_critic(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    projection_head: str = ASYMMETRIC_HEADS,
    output_dim: Optional[int] = None,
    nonlinearity=nn.ReLU(True),
    normalize: bool = False,
    temperature: float = 1.0,
):
    if not (projection_head in POSSIBLE_HEADS):
        raise ValueError(
            f"projection_heads must be one of {POSSIBLE_HEADS}, got {projection_head}"
        )

    if projection_head == ONE_HEAD:
        output_dim = y_dim
    else:
        if output_dim is None:
            raise ValueError("output_dim must be specified.")

    f_y = None

    assert len(hidden_dims) > 0
    f_x = DenseNN(
        input_dim=x_dim,
        hidden_dims=hidden_dims,
        param_dims=[output_dim],
        nonlinearity=nonlinearity,
    )
    if normalize:
        f_x = nn.Sequential(f_x, Normalize())

    if projection_head == SYMMETRIC_HEADS:
        f_y = f_x
    elif projection_head == ASYMMETRIC_HEADS:
        f_y = DenseNN(
            input_dim=y_dim,
            hidden_dims=hidden_dims,
            param_dims=[output_dim],
            nonlinearity=nonlinearity,
        )
        if normalize:
            f_y = nn.Sequential(f_y, Normalize())

    critic = SeparableCritic(
        f_x=f_x,
        f_y=f_y,
        temperature=temperature,
    )

    return critic


def joint_critic(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    nonlinearity=nn.ReLU(True),
):
    joint_net = DenseNN(
        input_dim=x_dim + y_dim,
        hidden_dims=hidden_dims,
        param_dims=[1],
        nonlinearity=nonlinearity,
    )
    unnormalized_log_ratio = JointCritic(
        joint_net=joint_net,
    )

    return unnormalized_log_ratio


def critic_nn(
    x_dim: int, y_dim: int, hidden_dims: List[int], critic_type: str, **kwargs
):
    if not (critic_type in CRITIC_TYPES):
        raise ValueError(
            f"critic must be one of {CRITIC_TYPES}, got {critic_type}"
        )

    if critic_type == JOINT_CRITIC:
        return joint_critic(
            x_dim=x_dim, y_dim=y_dim, hidden_dims=hidden_dims, **kwargs
        )
    else:
        return separable_critic(
            x_dim=x_dim, y_dim=y_dim, hidden_dims=hidden_dims, **kwargs
        )



from .functions import (
    QuantizationFunction,
    LearnableQuantization,
    FixedQuantization,
    VectorQuantization,
    LearnableVectorQuantization,
)
from .vqvae import VQVAE
from .factories import *



from typing import Dict, Tuple, Union
import torch


def unfold_samples(
    samples: Union[Dict[str, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]],
) -> Dict[str, torch.Tensor]:
    """
    Unify the format for the samples
    """
    variables = {}
    if isinstance(samples, tuple) or isinstance(samples, list):
        if not len(samples) == 2:
            raise Exception(
                "Dataloaders that iterate over tuples must have 2 elements, use dictionaries for more larger tuples"
            )
        x, y = samples
        variables["x"] = x
        variables["y"] = y
    elif isinstance(samples, dict):
        variables = samples
    else:
        raise NotImplementedError(
            "The dataloader must iterate over pairs or dictionaries"
        )

    if len(variables) < 2:
        raise Exception(
            f"Not enough variables to unfold, at least 2 are required."
        )

    return variables


def move_to_device(
    variables: Dict[str, torch.Tensor], device: torch.device
) -> Dict[str, torch.Tensor]:
    for name in variables:
        variables[name] = variables[name].to(device)
    return variables



from torch_mist.estimators.base import MIEstimator
from .generative.factories import *
from .discriminative.factories import *
from .transformed.factories import *


def instantiate_estimator(estimator_name, **kwargs) -> MIEstimator:
    factory_function = globals().get(estimator_name)

    # Check if the function name exists in the global namespace
    if not factory_function:
        raise ValueError(f"Factory function {estimator_name} not found.")
    estimator = factory_function(**kwargs)
    return estimator
